# -*- coding: utf-8 -*-

"""
OSSStorageDriver：
    * 列出Bucket中的Object
    * 获取Object的meta信息
    * 文件上传
    * 流式上传
    * 下载到本地文件
    * 流式下载
    * 复制
    * 删除
    * 批量删除
    * 归档
    * 恢复
    * 获取下载链接
统一使用crc64进行校验，不支持在request中自定义的``checksum_type``

"""

import datetime
import functools
import oss2
import oss2.exceptions as oss_exc

import gddriver.base as base
import gddriver.config as config
import gddriver.errors as err
import gddriver.models as models
import gddriver.utils.io as ioutil
import gddriver.utils.time as timeutil

DEFAULT_DOWNLOAD_BUFFER_SIZE = 1 << 14  # 16MB 默认下载时的缓冲大小
DEFAULT_SIZE_PER_PART = 1 << 20  # 1MB 默认每块分片的字节大小
DEFAULT_MULTIPART_TRANSFER_THRESHOLD = 1 << 26  # 64MB 默认分片传输的阈值
DEFAULT_URL_EXPIRES = 3600 * 12  # 12小时  下载链接默认过期时间
DEFAULT_LIST_OBJECTS = 100  # 默认遍历时的max_object
LIST_DATA_MAX_OBJECTS = 1000  # list_container_objects 时最大可获取的object数量
OSS_MAX_PART_NUM = 10000  # 分片上传最大分片数
PUT_DATA_MAX_SIZE = 5 * (1 << 30)  # 5GB  调用PutData api最大允许的文件大小

_DEFAULT_LOGGER_NAME = __name__


def exception_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger = config.get_logger(_DEFAULT_LOGGER_NAME)

        try:
            return func(*args, **kwargs)
        except oss_exc.NotFound as e:
            logger.debug("oss not found error", exc_info=1)
            details = e.details or {"error": str(e)}
            if e.code == oss_exc.NoSuchBucket.code:
                raise err.NoSuchContainer(details.get('BucketName'), **details)
            elif e.code == oss_exc.NoSuchKey.code:
                raise err.NoSuchObject(details.get('Key'), **details)
            else:
                raise err.NotFound(e.message, **details)
        except oss_exc.PositionNotEqualToLength as e:
            logger.debug("oss object position not equal to length", exc_info=True)
            details = e.details or {"error": str(e)}
            raise err.AppendPositionConflict(e.message, **details)
        except oss_exc.ServerError as e:
            # 4xx类型错误
            logger.debug("oss server error", exc_info=1)
            details = e.details or {"error": str(e)}
            if e.status == 400:
                raise err.BadRequest(e.message, code=e.code, **details)
            if e.status == 403:
                raise err.Forbidden(e.message, code=e.code, **details)
            else:
                raise err.DriverServerException(e.message, e.status, code=e.code, **details)
        except oss_exc.OssError as e:
            logger.debug("oss error", exc_info=1)
            status = e.status
            details = e.details or {"error": str(e)}
            if 'timeout' in str(e.body):
                raise err.RequestTimeout(e.body, status=status, **details)
            elif 'RequestError' in str(e.body):
                raise err.DriverRequestException(e.body, status=status, **details)
            else:
                raise err.StorageDriverException('unknown oss error: {}'.format(e), status=status, **details)

    return wrapper


class OSSConnection(base.Connection):
    def __init__(self, host, port, credential, timeout=None):
        super(OSSConnection, self).__init__(
            host=host,
            port=port,
            credential=credential
        )
        self.session = None
        self.connect_timeout = timeout

    def clone(self):
        return OSSConnection(
            host=self.host,
            port=self.port,
            credential=self.credential
        )

    def get_client(self, container_name):
        """
        获取OSS客户端 (Bucket实例)，可调用oss api

        :param container_name: 对应bucket_name
        :return: 返回bucket，在连接未关闭之前，会复用之前的连接池
        :rtype: :class:`oss2.Bucket`
        """
        access_key_token = self.credential.access_key_token
        access_key_id = self.credential.access_key_id
        access_key_secret = self.credential.access_key_secret

        if not access_key_token:
            auth = oss2.Auth(access_key_id, access_key_secret)
        else:
            auth = oss2.StsAuth(access_key_id, access_key_secret, access_key_token)

        # 在oss sdk中，如果传入session，则复用会话，否则新开始一个会话
        oss_bucket = oss2.Bucket(auth, self.host, container_name,
                                 session=self.session,
                                 connect_timeout=self.connect_timeout
                                 )
        self.session = self._get_session(oss_bucket)
        return oss_bucket

    def close(self):
        self.session = None

    @classmethod
    def _get_session(cls, oss_bucket):
        # 避免oss2模块升级后不提供session直接获取
        if hasattr(oss_bucket, 'session'):
            return oss_bucket.session
        return None


class OSSStorageDriver(base.StorageDriver):

    connection_class = OSSConnection
    checksum_type = 'crc'

    def __init__(self):
        super(OSSStorageDriver, self).__init__()

    @exception_handler
    def get_object_meta(self, connection, object_name, container_name):
        """
        获取object的详细信息
           *  name
           *  size
           *  extra
                * last_modified:    最后修改时间，ISO 8601格式
                * etag：            oss附带的etag信息
                * object_type:      对象的上传类型： Normal Multipart Appendable ...
                * storage_class:    对象的存储类型： Standard Archive ...
                * restore_status:   '' / restoring / restored
                * expiry_date:      归档文件恢复的过期时间，ISO 8601格式

        :param connection: `OSSConnection` instance
        :type  connection: :class:`OSSConnection`

        :param object_name: 对象名称
        :type  object_name: `str`

        :param container_name: bucket name
        :type  container_name: :class:`str`

        :return: 返回Object meta信息
        :rtype: :class:``gddriver.models.Object``
        """
        bucket = connection.get_client(container_name)
        meta = bucket.head_object(object_name)
        storage_class, restore_status, expiry_date = self.get_restore_status(meta)
        last_modified = datetime.datetime.fromtimestamp(meta.last_modified) if meta.last_modified else None
        extra = {
            "etag": meta.etag,
            "last_modified": last_modified,
            "object_type": meta.object_type,
            "storage_class": storage_class,
            "restore_status": restore_status,
            "expiry_date": expiry_date
        }
        crc = meta.resp.headers.get('x-oss-hash-crc64ecma', '')
        if crc:
            extra['crc64'] = crc

        # 不论正常返回200 OK还是非正常返回，HeadObject都不返回消息体
        # 如果文件类型为符号链接， 响应头中Content-Length、ETag、Content-Md5均为目标文件的元信息；
        # Last - Modified是目标文件和符号链接的最大值；其他均为符号链接元信息。
        # OSS服务根据协议RFC 1864对消息内容（不包括头部）计算MD5值获得128比特位数字，
        # 对该数字进行base64编码为一个消息的Content-MD5值
        md5 = meta.resp.headers.get('content-md5', '')
        if md5:
            extra['content-md5'] = md5
        return models.Object(object_name, meta.content_length, **extra)

    @exception_handler
    def list_container_objects(self, connection, container_name, prefix=None,
                               marker='', max_objects=DEFAULT_LIST_OBJECTS):
        """
        列出容器中的object

        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param container_name: bucket name
        :type  container_name: :class:`str`

        :param prefix: 文件的前缀（OSS中没有真正的目录的概念）
        :type  prefix: :class:`str`

        :param marker:  上一次遍历的标记，若marker不为None，则继续上一次标记遍历
        :type  marker: :class:`str`

        :param max_objects: 每次list返回的Object数目限制，最大可设置为1000
        :type  max_objects:  :class:`int`

        :return: object-meta列表, 下一次继续遍历的标记.
        :rtype: ``list`` of :class:`Object` | ``str``
        """
        _max_objects = max_objects if max_objects < LIST_DATA_MAX_OBJECTS else LIST_DATA_MAX_OBJECTS
        prefix = prefix or ''
        bucket = connection.get_client(container_name)
        result = bucket.list_objects(
            prefix=prefix,
            marker=marker,
            max_keys=_max_objects
        )
        is_truncated = result.is_truncated
        next_marker = None
        if is_truncated:
            next_marker = result.next_marker
        objects = []
        for meta in result.object_list:
            extra = {
                "etag": meta.etag,
                "last_modified": meta.last_modified,
                "object_type": meta.type,
                "storage_class": meta.storage_class
            }
            objects.append(models.Object(meta.key, meta.size, **extra))
        return objects, next_marker

    @exception_handler
    def download_object_as_stream(self, connection, request):
        """
        流式下载，获取一个可读的文件流
        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param request: 流式下载请求
        :type  request: gddriver.models.OSSStreamDownloadRequest

        :return: 返回的StreamDownloadResult是一个流，属性中默认包含从server端获取的crc
        :rtype: :class:``StreamDownloadResult``
        """
        logger = self.logger
        buffer_size = request.buffer_size
        container_name = request.container_name
        object_name = request.object_name
        progress_callback = request.progress_callback

        byte_range = request.byte_range if hasattr(request, 'byte_range') else None
        headers = request.headers if hasattr(request, "headers") else None

        buffer_size = buffer_size or DEFAULT_DOWNLOAD_BUFFER_SIZE
        object_meta = self.get_object_meta(connection, object_name, container_name)
        data_size = get_range_size(object_meta.size, byte_range)

        bucket = connection.get_client(container_name)
        stream = bucket.get_object(
            key=object_name,
            byte_range=byte_range,
            headers=headers
        )
        crc = stream.server_crc

        msg_template = ("download object as stream, bucket={}, object-key={}, "
                        "object-size={}, server-crc={}, buffer-size={}")
        msg = msg_template.format(container_name, object_name, object_meta.size, crc, buffer_size)
        logger.debug(msg)

        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        def generator():
            while True:
                chunk = stream.read(buffer_size)
                if not chunk:
                    break
                yield chunk
            logger.debug("%s, elapsed=%s, download finished", msg, stopwatch.elapsed())

        checksum_type = 'crc'
        stream_generator = generator()
        wrapped_stream = ioutil.make_progress_adapter(stream_generator, progress_callback, data_size)
        checksum_wrapped_stream = ioutil.make_checksum_adapter(wrapped_stream, checksum_type)
        result = models.StreamDownloadResult(
            iterator=checksum_wrapped_stream,
            length=data_size,
            server_checksum=crc,
            checksum_type=checksum_type
        )
        return result

    @exception_handler
    def download_file(self, connection, request):
        """
        下载文件到指定的路径中，支持断点续传

        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param request: 流式下载请求
        :type  request: gddriver.models.OSSDownloadRequest

        """

        container_name = request.container_name
        object_name = request.object_name
        local_file_path = request.file_path
        threads_count = request.threads_count
        progress_callback = request.progress_callback

        msg_template = ("multipart download, bucket={container_name}, object-key={object_name}, "
                        "dst-path={file_path}, threads_count={threads_count}")
        msg = msg_template.format(
            container_name=container_name,
            object_name=object_name,
            file_path=local_file_path,
            threads_count=threads_count
        )
        self.logger.debug(msg)

        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        bucket = connection.get_client(container_name)
        oss2.resumable_download(
            bucket=bucket,
            key=object_name,
            filename=local_file_path,
            num_threads=threads_count,
            progress_callback=progress_callback
        )
        self.logger.info("%s, elapsed=%s, download finished", msg, stopwatch.elapsed())

    @exception_handler
    def upload_object_via_stream(self, connection, request):
        """
        流式上传，可以上传网络流、文件流、字符串、bytes等可迭代的对象。

        :param connection: `OSSConnection` 实例
        :type  connection: :class:`OSSConnection`

        :param request:
        :type  request: `gddriver.models.OSSStreamUploadRequest`

        :rtype: :class:``gddriver.models.UploadResult``
        """

        container_name = request.container_name
        object_name = request.object_name
        stream = request.stream
        progress_callback = request.progress_callback

        data_size = request.data_size
        # 这是因为流式上传时，无法感知这个流的实际大小，默认会使用常规的put_object
        # 上传，但这种方式最大只能上传5GB文件，超过5GB必须使用分片上传，所以用户在知道
        # 文件流较大时，可以使用multi参数通知driver使用分片上传
        multi = request.multi if hasattr(request, 'multi') else False

        logger = self.logger
        msg_template = ("upload via stream, bucket={container_name}, object-key={object_name}, "
                        "checksum-type={checksum_type}, data-size={data_size}, multi={multi}")
        msg = msg_template.format(
            container_name=container_name,
            object_name=object_name,
            checksum_type=self.checksum_type,
            data_size=data_size,
            multi=multi
        )
        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        logger.debug(msg)

        wrapped_stream = ioutil.make_progress_adapter(stream, progress_callback, data_size)

        bucket = connection.get_client(container_name)
        if not multi:
            oss_result = bucket.put_object(object_name, wrapped_stream)
        else:
            # 分片上传
            logger.debug("%s, via multipart mode", msg)
            upload_id = bucket.init_multipart_upload(object_name).upload_id
            parts = []
            part_num = 0
            for chunk in wrapped_stream:
                part_num += 1
                result = bucket.upload_part(object_name, upload_id, part_num, chunk)
                parts.append(oss2.models.PartInfo(part_num, result.etag))
            oss_result = bucket.complete_multipart_upload(object_name, upload_id, parts)

        server_checksum = oss_result.crc
        extra = self.__get_extra_info(oss_result)
        result = models.UploadResult(
            checksum=server_checksum,
            checksum_type=self.checksum_type,
            **extra
        )
        logger.info('%s, checksum=%s, elapsed=%s, upload finished',
                    msg, result.checksum, stopwatch.elapsed())
        return result

    @exception_handler
    def upload_file(self, connection, request):
        """
        分片多线程上传一个文件

        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param request:
        :type  request: `gddriver.models.OSSUploadRequest`

        :return:
        :rtype `UploadResult`
        """
        container_name = request.container_name
        file_path = request.file_path
        object_name = request.object_name
        threads_count = request.threads_count
        progress_callback = request.progress_callback

        part_size = request.part_size if hasattr(request, 'part_size') else None
        headers = request.headers if hasattr(request, 'headers') else None

        logger = self.logger
        msg_template = ("multipart upload, file={file_path}, bucket={container_name}, object-key={object_name}, "
                        "threads_count={threads_count}, part-size={part_size}, checksum-type={checksum_type}")
        msg = msg_template.format(
            file_path=file_path,
            container_name=container_name,
            object_name=object_name,
            threads_count=threads_count,
            part_size=part_size,
            checksum_type=self.checksum_type
        )
        logger.debug(msg)

        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        bucket = connection.get_client(container_name)
        oss_result = oss2.resumable_upload(
            bucket=bucket,
            key=object_name,
            filename=file_path,
            headers=headers,
            part_size=part_size,
            progress_callback=progress_callback,
            num_threads=threads_count
        )

        server_checksum = oss_result.crc
        extra = self.__get_extra_info(oss_result)
        result = models.UploadResult(
            checksum=server_checksum,
            checksum_type=self.checksum_type,
            **extra
        )
        logger.info('{}, checksum={}, elapsed={}, upload finished'.format(
            msg, server_checksum, stopwatch.elapsed())
        )
        return result

    @exception_handler
    def append_object(self, connection, request):
        """
        追加上传，可以将网络流、文件流等追加到指定的object_name, 最大不能超过5G

        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param request:
        :type  request: `gddriver.models.AppendRequest`

        :return: 追加上传的结果 （返回的校验码为追加後整个文件的校验码）
        :rtype: :class:``gddriver.models.AppendResult``
        """

        container_name = request.container_name
        object_name = request.object_name
        stream = request.stream
        progress_callback = request.progress_callback
        position = request.position
        data_size = request.data_size

        logger = self.logger
        msg_template = ("append object, bucket={container_name}, object-name={object_name}, "
                        "position={position}, data-size={data_size}, checksum-type={checksum_type}")
        msg = msg_template.format(
            container_name=container_name,
            object_name=object_name,
            position=position,
            data_size=data_size,
            checksum_type=self.checksum_type
        )
        logger.debug(msg)
        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        bucket = connection.get_client(container_name)
        progress_wrapped_stream = ioutil.make_progress_adapter(stream, progress_callback, data_size)

        init_crc = 0
        if position != 0:
            try:
                object_meta = self.get_object_meta(
                    connection=connection,
                    container_name=container_name,
                    object_name=object_name
                )
                init_crc = object_meta.extra.get('crc64', None)
            except err.NoSuchObject:
                # 用户追加上传到一个不存在的文件时，但position不为0，捕获异常，oss重新抛出後由装饰器统一处理
                logger.debug('append object not found: %s', msg)
        oss_result = bucket.append_object(object_name, position, progress_wrapped_stream, init_crc=init_crc)
        server_checksum = oss_result.crc

        next_position = oss_result.next_position
        extra = self.__get_extra_info(oss_result)
        result = models.AppendResult(
            next_position=next_position,
            checksum=server_checksum,
            checksum_type=self.checksum_type,
            **extra
        )
        logger.info("%s, checksum=%s, elapsed=%s, upload finished",
                    msg, server_checksum, stopwatch.elapsed())
        return result

    @exception_handler
    def delete_object(self, connection, object_name, container_name=None):
        bucket = connection.get_client(container_name)
        bucket.delete_object(object_name)

    @exception_handler
    def copy_object(self, connection, request):
        """
        复制文件
          复制文件较大时，必须调整part_size从而保证总分片数不超过10000 (OSS限制)

          OSS相关限制条件：
            请求者必须对源Object有读权限。
            源Object和目标Object必须属于同一个地域（数据中心）。
            不能拷贝通过追加上传方式产生的Object。
            如果源Object为符号链接，只拷贝符号链接，不拷贝符号链接指向的文件内容。

        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param request: 对象复制请求
        :type  request: :class:``gddriver.models.OSSCopyRequest``
        :return:
        """

        logger = self.logger

        src_container_name = request.container_name
        src_object_name = request.object_name
        # 默认复制到当前容器中
        dst_container_name = request.dst_container_name or src_container_name
        dst_object_name = request.dst_object_name

        # 兼容基类的参数
        headers = request.headers if hasattr(request, 'headers') else None
        multi_threshold = request.multi_threshold if hasattr(request, 'multi_threshold') else None
        final_threshold = multi_threshold or DEFAULT_MULTIPART_TRANSFER_THRESHOLD

        object_meta = self.get_object_meta(connection, src_object_name, src_container_name)

        msg_template = ("copy object, from={src_container_name}:{src_object_name}, "
                        "to={dst_container_name}:{dst_object_name}, size={size}")
        msg = msg_template.format(
            src_container_name=src_container_name,
            src_object_name=src_object_name,
            dst_container_name=dst_container_name,
            dst_object_name=dst_object_name,
            size=object_meta.size
        )
        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        dst_bucket = connection.get_client(dst_container_name)
        if object_meta.size <= final_threshold:
            logger.debug(msg)
            dst_bucket.copy_object(
                source_bucket_name=src_container_name,
                source_key=src_object_name,
                target_key=dst_object_name,
                headers=headers
            )
        else:
            # 文件较大时，使用分片复制
            upload_id = dst_bucket.init_multipart_upload(dst_object_name).upload_id
            part_size = regulate_part_size(object_meta.size)

            logger.debug("%s, multipart copy, upload-id=%s, part-size=%s", msg, upload_id, part_size)
            parts = []
            part_number = 1
            offset = 0
            while offset < object_meta.size:
                parts_to_upload = min(part_size, object_meta.size - offset)
                # 左闭右闭
                byte_range = (offset, offset + parts_to_upload - 1)
                part_copy_result = dst_bucket.upload_part_copy(
                    source_bucket_name=src_container_name,
                    source_key=src_object_name,
                    byte_range=byte_range,
                    target_key=dst_object_name,
                    target_upload_id=upload_id,
                    target_part_number=part_number,
                    headers=headers
                )
                parts.append(oss2.models.PartInfo(part_number, part_copy_result.etag))
                offset += parts_to_upload
                part_number += 1
            logger.info("%s %s copied  elapsed=%s, finished", msg, offset, stopwatch.elapsed())
            dst_bucket.complete_multipart_upload(dst_object_name, upload_id, parts)

    @exception_handler
    def batch_delete_objects(self, connection, object_name_list, container_name=None):
        """批量删除文件。

        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param container_name: bucket name
        :type  container_name: :class:`str`

        :param object_name_list:
        :type object_name_list: list of str

        """
        logger = self.logger
        logger.debug("batch delete objects, %s", object_name_list)
        if not object_name_list:
            logger.warning("object key list is None.")
            return
        # TODO 确认批量删除部分失败的场景
        bucket = connection.get_client(container_name)
        batch_delete_result = bucket.batch_delete_objects(object_name_list)
        logger.info("batch delete objects, result=%s", batch_delete_result)

    @exception_handler
    def archive_object(self, connection, src_container_name, src_object_name,
                       archive_container_name, archive_object_name, delete=False):
        """
        将文件从一个bucket中转归档到另一个bucket中，并删除原来的文件，适合不同类型的bucket之间的复制

        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param src_container_name: 原数据的bucket name
        :param src_object_name: 原数据的object_key
        :param archive_container_name: 要归档至的bucket name
        :param archive_object_name: 归档后的object key
        :param delete: 是否在归档完成后删除原文件
        """
        self.__download_and_upload_between_buckets(
            connection=connection,
            src_bucket_name=src_container_name,
            src_obj_key=src_object_name,
            dst_bucket_name=archive_container_name,
            dst_obj_key=archive_object_name
        )
        if delete:
            self.delete_object(
                connection=connection,
                container_name=src_container_name,
                object_name=src_object_name
            )

    @exception_handler
    def restore_object(self, connection, src_container_name, src_object_name):
        """
        将低频存储类型暂时恢复为可访问的类型
        :param connection: `OSSConnection`
        :param src_container_name: 原数据的bucket name
        :param src_object_name: 原数据的object_key
        :rtype: :class:``gddriver.models.RestoreResult``
        """

        archive_bucket = connection.get_client(src_container_name)
        try:
            self.logger.debug('restore: %s start', src_object_name)
            rsp = archive_bucket.restore_object(src_object_name)
            self.logger.debug('restore: %s finished, response status: %s', src_object_name, rsp.status)
            if rsp.status == 202:
                return models.RestoreResult(finished=False)
        except oss_exc.RestoreAlreadyInProgress:
            self.logger.info("restore already in process: %s/%s", src_container_name, src_object_name)
            return models.RestoreResult(finished=False)

        meta = archive_bucket.head_object(src_object_name)
        _, restore_status, expiry_date = self.get_restore_status(meta)
        finished = restore_status == 'restored'
        return models.RestoreResult(
            finished=finished,
            expiry_date=expiry_date
        )

    @exception_handler
    def move_object(self, connection, src_object_name, dst_object_name, container_name=None, dst_container_name=None):
        """
        将object移动到另一路径（不支持从标准存储向低频存储中移动，如有需要，则使用archive_object），

        :param connection: `OSSConnection`
        :type  connection: :class:`OSSConnection`

        :param container_name: 原数据的bucket name
        :param src_object_name: 原数据的object_key
        :param dst_object_name: 要归档至的bucket name
        :param dst_container_name: 归档后的object key
        """
        copy_request = models.OSSCopyRequest(
            container_name=container_name,
            object_name=src_object_name,
            dst_container_name=dst_container_name or container_name,
            dst_object_name=dst_object_name
        )

        object_meta = self.get_object_meta(
            connection=connection,
            container_name=copy_request.container_name,
            object_name=copy_request.object_name
        )
        if object_meta.extra.get('object_type') == 'Appendable':
            # Append类型的文件move之后仍然是Append类型
            self._copy_append_type_object(
                connection=connection,
                copy_request=copy_request,
                object_meta=object_meta
            )
        else:
            self.copy_object(
                connection=connection,
                request=copy_request
            )
        self.delete_object(
            connection=connection,
            container_name=container_name,
            object_name=src_object_name
        )

    @exception_handler
    def get_object_sign_url(self, connection, object_name, container_name=None, expires=DEFAULT_URL_EXPIRES,
                            headers=None):
        """

        :param connection:
        :param object_name:
        :param container_name:
        :param headers: 需要签名的HTTP头部，如名称以x-oss-meta-开头的头部（作为用户自定义元数据）、
                        Content-Type头部等。对于下载，不需要填。
        :param expires:  过期时间 (秒)
        :return:
        """
        bucket = connection.get_client(container_name)
        expires = expires or DEFAULT_URL_EXPIRES
        return bucket.sign_url('GET', object_name, expires=expires, headers=headers)

    @exception_handler
    def init_multipart_upload(self, connection, object_name, container_name=None, overwrite=False):
        """
        :param connection:
        :type connection: OSSConnection
        :param object_name:
        :param container_name:
        :param overwrite: 默认直接覆盖上传，如非覆盖，则抛出文件已存在的异常
        :return:
        """

        if not overwrite:
            try:
                self.get_object_meta(
                    container_name=container_name,
                    connection=connection,
                    object_name=object_name
                )
                raise err.FileAlreadyExists("{} already exists".format(object_name))
            except err.NotFound:
                pass

        bucket = connection.get_client(container_name)
        result = bucket.init_multipart_upload(object_name)
        self.logger.debug(result)
        return result.upload_id

    @exception_handler
    def upload_part(self, connection, request):

        bucket = connection.get_client(request.container_name)
        oss_response = bucket.upload_part(
            key=request.object_name,
            upload_id=request.upload_id,
            part_number=request.part_number,
            data=request.stream,
            progress_callback=request.progress_callback
        )
        return models.PartUploadResult(
            checksum=oss_response.crc,
            checksum_type='crc',
            part_info=models.PartInfo(part_number=request.part_number, etag=oss_response.etag)
        )

    @exception_handler
    def complete_multipart_upload(self, connection, request):

        def map_to_oss_part_info(part):
            """
            :type part: gddriver.models.PartInfo
            """
            return oss2.models.PartInfo(part.part_number, part.etag)

        bucket = connection.get_client(request.container_name)
        parts = map(map_to_oss_part_info, request.parts)
        models.PartInfo.sort_parts(list(parts))
        oss_result = bucket.complete_multipart_upload(request.object_name, request.upload_id, parts)
        self.logger.debug(oss_result)

    def _copy_append_type_object(self, connection, copy_request, object_meta=None):
        """
        object type 为Append 的情况下，无法直接使用copy，需通过流式传输复制文件

        :type connection: OSSConnection
        :type copy_request: gddriver.models.CopyRequest
        """

        self.logger.debug("copy APPEND object from %s to %s, src container: %s, dst container: %s",
                          copy_request.object_name, copy_request.dst_object_name,
                          copy_request.container_name, copy_request.dst_container_name)

        if not object_meta:
            object_meta = self.get_object_meta(
                connection=connection,
                container_name=copy_request.container_name,
                object_name=copy_request.object_name
            )

        object_size = object_meta.size
        part_size = regulate_part_size(object_size)

        stream_download_request = models.OSSStreamDownloadRequest(
            container_name=copy_request.container_name,
            object_name=copy_request.object_name
        )
        stream_download_request.buffer_size = part_size

        stream = self.download_object_as_stream(
            connection=connection,
            request=stream_download_request
        )

        append_request = models.AppendRequest(
            container_name=copy_request.dst_container_name,
            object_name=copy_request.dst_object_name,
            stream=stream,
            position=0
        )
        result = self.append_object(
            connection=connection,
            request=append_request
        )
        self.logger.debug("copy APPEND object from %s to %s, src container: %s, dst container: %s, result: %s",
                          copy_request.object_name, copy_request.dst_object_name,
                          copy_request.container_name, copy_request.dst_container_name, result)
        self.logger.debug("copy")

    def __download_and_upload_between_buckets(self, connection, src_bucket_name, src_obj_key,
                                              dst_bucket_name, dst_obj_key):

        # TODO 50GB的文件相当于每个分块至少5MB，并发量大时对性能消耗比较大

        msg = ("copy between buckets, from={src_bucket_name}:{src_obj_key}, "
               "to={dst_bucket_name}:{dst_obj_key}").format(
            src_bucket_name=src_bucket_name,
            src_obj_key=src_obj_key,
            dst_bucket_name=dst_bucket_name,
            dst_obj_key=dst_obj_key
        )
        self.logger.debug(msg)

        object_meta = self.get_object_meta(
            connection=connection,
            container_name=src_bucket_name,
            object_name=src_obj_key
        )

        object_size = object_meta.size
        multi_part_upload = object_size > PUT_DATA_MAX_SIZE
        part_size = regulate_part_size(object_size)

        msg = "{}, size={}, chunk-size={}".format(msg, object_size, part_size)
        self.logger.debug(msg)
        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        stream_download_request = models.OSSStreamDownloadRequest(
            container_name=src_bucket_name,
            object_name=src_obj_key
        )
        stream_download_request.buffer_size = part_size

        stream = self.download_object_as_stream(
            connection=connection,
            request=stream_download_request
        )

        stream_upload_request = models.OSSStreamUploadRequest(
            container_name=dst_bucket_name,
            stream=stream,
            object_name=dst_obj_key
        )
        stream_upload_request.multi = multi_part_upload

        result = self.upload_object_via_stream(
            connection=connection,
            request=stream_upload_request
        )
        self.logger.info("%s, result=%s, elapsed=%s, copy finished", msg, result, stopwatch.elapsed())
        return result

    @classmethod
    def _verify_and_get_server_checksum(cls, message, oss_result, client_checksum):
        """验证校验码，并将服务端校验码返回"""
        server_checksum = oss_result.crc
        if server_checksum != client_checksum:
            raise err.ChecksumMismatch(
                message=message,
                server_checksum=server_checksum,
                client_checksum=client_checksum
            )
        return server_checksum

    @staticmethod
    def __get_extra_info(result):
        """
        :param result: oss api请求的返回值
        :rtype: dict
        """
        return {'request_id': result.request_id, 'etag': result.etag}

    @classmethod
    def get_restore_status(cls, meta):
        """
        通过meta获取restore的状态信息
        :param meta:
        :type meta: :class:``oss2.models.HeadObjectResult``
        :return: 存储类型， restore状态， 过期时间
        :rtype: str | str | datetime.datetime
        """
        storage_class, restore, finished, expiry_date = cls.__get_restored_info(meta)
        # 普通文件
        if storage_class != oss2.BUCKET_STORAGE_CLASS_ARCHIVE:
            return storage_class, '', None
        # archived
        elif not restore:
            return storage_class, 'archived', None
        # restoring or restored
        elif finished:
            return storage_class, 'restored', expiry_date
        else:
            return storage_class, 'restoring', None

    @classmethod
    def __get_restored_info(cls, meta):
        """
        当文件非restoring或restored状态的文件，返回的第二个参数restore 为False，
        is_finished 为False，expiry_date 为None

        :param meta: meta of head_object
        :return: storage_class, is_restore_object, is_finished_restore, expiry
        """
        finished = False
        restore = False
        expiry_date = None

        # 如果Bucket类型为Archive，且用户已经提交Restore请求，则响应头中会以x-oss-restore返回该Object的Restore状态，分如下几种情况：
        #     a.如果没有提交Restore或者Restore已经超时，则不返回该字段。
        #     b.如果已经提交Restore，且Restore没有时完成，则返回的x-oss-restore值为: ongoing-request ="true"。
        #     c.如果已经提交Restore，且Restore已经完成，
        #       则返回的x-oss-restore值为: ongoing-request ="false", expiry-date ="Sun, 16 Apr 2017 08:12:33 GMT"。

        # x-oss-storage-class展示Object的存储类型：Standard，IA，Archive
        storage_class = meta.resp.headers['x-oss-storage-class']
        if storage_class != oss2.BUCKET_STORAGE_CLASS_ARCHIVE:
            return storage_class, restore, finished, expiry_date

        # 状态为restoring的文件，meta中包含信息 'x-oss-restore'为True，
        # 同时也可以从meta.resp.headers中获取： meta.resp.headers['x-oss-restore'] == 'ongoing-request="true"':
        restore_progress = meta.headers.get('x-oss-restore', '')
        if not restore_progress:
            return storage_class, restore, finished, expiry_date
        else:
            restore = True

        if 'true' not in restore_progress:
            # x-oss-restore: ongoing-request="false", expiry-date="Sun, 16 Apr 2017 08:12:33 GMT"
            finished = True
            restore = True
            expiry_time = cls.get_expiry_time(restore_progress)
            expiry_date = timeutil.timestamp_to_date(expiry_time)

        return storage_class, restore, finished, expiry_date

    @classmethod
    def get_expiry_time(cls, restore_progress):
        """Get restore expiry time from restore progress string

        :param restore_progress: e.g. ongoing-request="false", expiry-date="Sun, 16 Apr 2017 08:12:33 GMT"
        :return Unix 时间戳
        """
        expiry_date_str = restore_progress.split(',', 1)[1]
        expiry_date_str = expiry_date_str.replace(' expiry-date="', '').replace('"', '')
        expiry_time = oss2.utils.http_to_unixtime(expiry_date_str)
        return expiry_time


def regulate_part_size(object_size):
    part_size = DEFAULT_SIZE_PER_PART
    while part_size * OSS_MAX_PART_NUM < object_size:
        # part_size 按MB递增
        part_size += part_size
    return part_size


def get_range_size(object_size, byte_range):
    """下载范围: (left, right) 来自oss2注释
        - byte_range 为 (0, 99)  -> 'bytes=0-99'，表示读取前100个字节
        - byte_range 为 (None, 99) -> 'bytes=-99'，表示读取最后99个字节
        - byte_range 为 (100, None) -> 'bytes=100-'，表示读取第101个字节到文件结尾的部分（包含第101个字节）
    """
    if not byte_range:
        return object_size
    left = byte_range[0]
    right = byte_range[1]
    if left == 0 and right > 0:
        return right - left + 1
    elif left is None and right > 0:
        return right
    elif left >= 0 and right is None:
        return object_size - left
    elif 0 <= left < right:
        return right - left + 1
    else:
        raise ValueError("illegal byte range")



# -*- coding: utf-8 -*-


__all__ = ['Provider', 'get_driver']


class Provider(object):
    OSS = 'oss'
    FTP = 'ftp'


_drivers = {
    Provider.OSS: ("gddriver.drivers.oss", "OSSStorageDriver"),
    Provider.FTP: ("gddriver.drivers.ftp", "FTPStorageDriver")
}


def get_driver(provider):
    """

    :param provider:
    :rtype: :class:`gddriver.base.StorageDriver`
    """

    if provider in _drivers:
        module_name, driver_name = _drivers[provider]
        module = __import__(module_name, globals(), locals(), [driver_name])
        try:
            klass = getattr(module, driver_name)
            return klass()
        except AttributeError:
            raise ImportError('Not implemented for this driver: {}'.format(provider))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals


class StorageDriverException(Exception):
    message = "Storage driver operation exception."

    def __init__(self, message=None, **kwargs):
        self.message = message or self.message
        self.details = {}
        self.details.update(**kwargs)

    def __str__(self):
        return "{}  details: {}".format(self.message, self.details or None)


class DriverLocalException(StorageDriverException):
    message = "Storage driver local operation exception."


class DriverRequestException(StorageDriverException):
    message = "Storage driver request operation exception."


class DriverServerException(StorageDriverException):
    message = "Storage driver server process exception."
    status = 500

    def __init__(self, message=None, status=None, **kwargs):
        super(DriverServerException, self).__init__(message, **kwargs)
        self.status = status or self.status

    def __str__(self):
        return '{} {}'.format(self.status, super(DriverServerException, self).__str__())


class RequestTimeout(DriverRequestException):
    message = "Request timeout"


class IllegalParameter(DriverLocalException):
    message = "Illegal parameter"

    def __init__(self, name=None, value=None):
        if name:
            self.message += ", " + name
        if value:
            self.message = "{}: {}".format(self.message, value)
        super(IllegalParameter, self).__init__(self.message)


class DataAdapterWrapException(DriverLocalException):

    def __init__(self, data):
        self.message = '{0} is not a file object, nor an iterator'.format(data.__class__.__name__)
        super(DataAdapterWrapException, self).__init__(self.message)


class FileAlreadyExists(DriverLocalException):
    message = "File already exists"

    def __init__(self, msg):
        super(FileAlreadyExists, self).__init__(msg)


class PathTypeConflict(DriverLocalException):
    message = "Path conflict."


class TransferError(DriverLocalException):
    message = "Transfer error."


class UnexpectedDownloadedSize(TransferError):
    def __init__(self, expect_size, actual_size):
        self.message = "Unexpected downloaded file size: expect: {}, actual: {}".format(expect_size, actual_size)
        super(UnexpectedDownloadedSize, self).__init__(self.message)


class GetConnectException(DriverLocalException):
    message = "Get connect exception, miss parameters."


class BadRequest(DriverServerException):
    status = 400


class OperationNotSupported(BadRequest):
    message = "Operation not supported"


class Forbidden(DriverServerException):
    status = 403
    message = "Forbidden"


class NotLoggedIn(Forbidden):
    pass


class NotFound(DriverServerException):
    status = 404
    message = "Not found"


class NoSuchObject(NotFound):
    def __init__(self, object_name, **kwargs):
        self.message = "No such object: {}".format(object_name)
        super(NoSuchObject, self).__init__(**kwargs)


class NoSuchContainer(NotFound):
    def __init__(self, container_name, **kwargs):
        self.message = "No such container: {}".format(container_name)
        super(NoSuchContainer, self).__init__(**kwargs)


class Conflict(DriverServerException):
    status = 409
    message = 'conflict'


class AppendPositionConflict(Conflict):
    def __init__(self, position, **kwargs):
        self.message = "Position is not equal to remote file length {}".format(position)
        super(AppendPositionConflict, self).__init__(**kwargs)


class ChecksumMismatch(Conflict):
    def __init__(self, message, server_checksum, client_checksum, **kwargs):
        self.message = "{}, checksum mismatch: server checksum: {}, client checksum: {}".format(
            message, server_checksum, client_checksum
        )
        super(ChecksumMismatch, self).__init__(**kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import collections
import hashlib
import os
import threading
import struct

import crcmod

import gddriver.utils as utils
import gddriver.errors as gderrors

Iterable = collections.Iterable
_CHUNK_SIZE = 1024 * 100


"""
str_type： 不可直接作为bytes使用的类型，进行md5、crc等计算前需要encode操作

对于PY2：str 等价于 bytes，不需要编码操作
对于PY3：str 不同于 bytes，需要显式地编码操作
"""
if utils.PY2:
    str_type = (unicode,)  # py2 - unicode: unencoded，builtin str: encoded
else:
    str_type = (str,)  # py3 - builtin str: unencoded


def make_checksum_adapter(data, checksum_type='crc'):
    """为data添加checksum/checksum_type属性"""
    if hasattr(data, 'read'):
        return _FileLikeAdapter(data, checksum_type=checksum_type)
    elif isinstance(data, Iterable):
        return _IterableAdapter(data, checksum_type=checksum_type)
    else:
        raise gderrors.DataAdapterWrapException(data)


def make_progress_adapter(data, progress_callback, data_size=None):
    """为data添加进度回调"""

    if hasattr(data, 'read'):
        return _FileLikeAdapter(data, progress_callback=progress_callback, data_size=data_size)
    elif isinstance(data, Iterable):
        return _IterableAdapter(data, progress_callback=progress_callback, data_size=data_size)
    else:
        raise gderrors.DataAdapterWrapException(data)


def make_file_like_adapter(data):
    """将普通的iterator转换为FileLike对象（增加read操作）"""
    if hasattr(data, 'read'):
        return data
    elif isinstance(data, Iterable):
        return _IteratorToFileAdapter(data)
    else:
        raise gderrors.DataAdapterWrapException(data)


def make_file_iterator(file_path, chunk_size=8092):
    with open(file_path, "rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            yield chunk


class Adapter(object):
    """progress_callback/checksum 适配器，为对象添加progress_callback功能/checksum属性"""
    def __init__(self, progress_callback=None, checksum_type=None, data_size=None, **kwargs):

        self.progress_callback = progress_callback
        self._data_size = data_size
        self._checksum_callback = None
        self.__checksum_type = checksum_type
        if checksum_type:
            if checksum_type.lower() == 'crc':
                self._checksum_callback = Crc64()
            elif checksum_type.lower() == 'md5':
                self._checksum_callback = MD5()

    @property
    def checksum(self):
        if not self._checksum_callback:
            return None
        return self._checksum_callback.checksum

    @property
    def checksum_type(self):
        return self.__checksum_type


class _FileLikeAdapter(Adapter):
    """
    checksum/progress_callback适配器，包装带有'read'方法的可迭代类型，
    为其添加progress_callback功能/checksum属性
    """

    def __init__(self, data, progress_callback=None, checksum_type=None, data_size=None):
        super(_FileLikeAdapter, self).__init__(progress_callback, checksum_type, data_size)

        # encode to bytes (utf-8)
        self.file_data = str_to_bytes(data)
        self._data_size = data_size or _get_data_size(data)
        self.offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        content = self.read(_CHUNK_SIZE)

        if content:
            return content
        else:
            raise StopIteration

    def read(self, amt=None):
        content = self.file_data.read(amt)
        if not content:
            _progress_callback(self.progress_callback, self.offset, self._data_size)
        else:
            _progress_callback(self.progress_callback, self.offset, self._data_size)

            self.offset += len(content)

            _checksum_callback(self._checksum_callback, content)
        return content


class _IterableAdapter(Adapter):
    """
     checksum/progress_callback适配器，包装普通的可迭代类型，
    为其添加progress_callback功能/checksum属性
    """
    def __init__(self, data, progress_callback=None, checksum_type=None, data_size=None):
        super(_IterableAdapter, self).__init__(progress_callback, checksum_type, data_size)

        # encode to bytes (utf-8)
        encoded_data = str_to_bytes(data)
        if type(data) == bytes or type(data) == str:
            # bytes to iterator
            self.iter = iter([encoded_data])
        else:
            self.iter = iter(encoded_data)
        self.offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        _progress_callback(self.progress_callback, self.offset, self._data_size)

        content = next(self.iter)
        self.offset += len(content)

        _checksum_callback(self._checksum_callback, content)

        return content


class _IteratorToFileAdapter(object):
    """
    FileLike适配器，将普通的iterator转换为FileLike对象
    """
    def __init__(self, iterator):
        self.iter = iterator
        self.remain = None
        self.offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):

        content = self.read(_CHUNK_SIZE)

        if content:
            return content
        else:
            raise StopIteration

    def read(self, amt=None):
        amt = int(amt or 0)
        content = self._read_with_amt(amt) if amt else self._read_without_amt()
        if content:
            self.offset += len(content)
        return content

    def _read_without_amt(self):
        content = b''
        content += self.remain or b''
        self.remain = None
        while True:
            try:
                data = unpack_to_bytes(next(self.iter))
                content += data
            except StopIteration:
                break
        return content

    def _read_with_amt(self, amt):
        content = b''
        if self.remain:
            buf = min(amt, len(self.remain))
            content = self.remain[:buf]
            self.remain = self.remain[buf:]

        while not content or len(content) < amt:
            need = amt - len(content)
            try:
                self.remain = unpack_to_bytes(next(self.iter))
            except StopIteration:
                break
            content += self.remain[:need]
            self.remain = self.remain[need:]
        return content


def _progress_callback(function, offset, data_size):
    if function:
        function(offset, data_size)


def _checksum_callback(checksum_callback, content):
    if checksum_callback:
        checksum_callback(content)


def _get_data_size(data):
    if hasattr(data, '__len__'):
        return len(data)

    if hasattr(data, 'seek') and hasattr(data, 'tell'):
        return file_object_remaining_bytes(data)

    return None


def file_object_remaining_bytes(fileobj):
    current = fileobj.tell()

    fileobj.seek(0, os.SEEK_END)
    end = fileobj.tell()
    fileobj.seek(current, os.SEEK_SET)

    return end - current


class Crc64(object):
    """
    使用与oss相同的生成多项式(ECMA-182标准)计算crc64
    """
    _POLY = 0x142F0E1EBA9EA3693
    _XOROUT = 0xFFFFFFFFFFFFFFFF

    def __init__(self, init_crc=0):
        self.crc64 = crcmod.Crc(self._POLY, initCrc=init_crc, rev=True, xorOut=self._XOROUT)

    def __call__(self, data):
        self.update(data)

    def update(self, data):
        self.crc64.update(data)

    @property
    def checksum(self):
        return self.crc64.crcValue


class MD5(object):

    def __init__(self):
        self.md5 = hashlib.md5()

    def __call__(self, data):
        self.update(data)

    def update(self, data):
        self.md5.update(data)

    @property
    def checksum(self):
        return self.md5.hexdigest()


def copy_file(stream, dst):
    size = 0
    for chunk in stream:
        dst.write(chunk)
        size += len(chunk)
    return size


def create_checksum_yield(file_path, checksum_type, logger=None):
    f = open(file_path, 'rb')
    adapter = make_checksum_adapter(f, checksum_type)

    def calculate():
        if logger:
            logger.debug("DriverAction=checksum, file_path={}".format(file_path))
        try:
            if not checksum_type:
                return
            for ignored in adapter:
                pass
        finally:
            f.close()

    task = threading.Thread(target=calculate, args=())
    task.start()

    def generator():
        task.join()
        yield adapter.checksum

    return generator()


def str_to_bytes(data):
    """若输入为str（unicode），则转为utf-8编码的bytes；其他则原样返回"""
    if isinstance(data, str_type):
        return data.encode(encoding='utf-8')
    else:
        return data


def unpack_to_bytes(data):
    """
    将迭代器中获取的内容转换为bytes内容:
        str -> bytes
        int -> bytes
    :param data: str/bytes/int
    :return: bytes
    """

    content = str_to_bytes(data)

    if isinstance(content, int):
        return struct.pack('B', content)
    return content



# -*- coding: utf-8 -*-

import gddriver.utils.io as io_util
__all__ = [
    'Object',
    'Credential',
    'StreamTransferRequest',
    'FTPStreamUploadRequest',
    'OSSStreamUploadRequest',
    'AppendRequest',
    'StreamDownloadRequest',
    'DownloadResult',
    'StreamDownloadResult',
    'OperationResult',
    'UploadResult',
    'AppendResult',
    'RestoreResult',
    'CopyRequest',
    'OSSCopyRequest'
]


class Object(object):
    def __init__(self, name, size=0, **kwargs):
        """
        对象元信息

        :param name: 对象名称 (必须是在容器中唯一的).
        :type  name: ``str``

        :param size: 对象大小（单位：字节）
        :type  size: ``int``

        :param kwargs: 对象的拓展信息，记录到extra中，根据不同的存储服务有不同的关键字

        extra
            + oss:
                * last_modified:    UNIX TIMPSTAMP
                * etag：            etag
                * object_type:      Normal Multipart Appendable ...
                * storage_class:    Standard Archive ...
                * restore:          是否为恢复/恢复中的归档文件 （get_object_meta时获得）
                * restore_finished: 是否恢复完成  （get_object_meta时获得）
                * expiry_date:      归档文件恢复的过期时间 （get_object_meta时获得）
            + ftp:
                * parent:         /root/prefix
                * is_file:        是否为文件
                * short_name:     对象名，没有前缀
        """
        self.name = name
        self.extra = {}
        self.size = size
        if kwargs:
            self.extra.update(**kwargs)

    @property
    def storage_class(self):
        return self.extra.get('storage_class', 'Standard')

    def __str__(self):
        return 'name: {name}, size: {size}, extra: {extra}'.format(
            name=self.name,
            size=self.size,
            extra=self.extra
        )


class Credential(object):
    def __init__(self, access_key_id=None, access_key_secret=None, access_key_token=None,
                 user=None, password=None):
        self._access_key_id = access_key_id
        self._access_key_secret = access_key_secret
        self._access_key_token = access_key_token
        self._user = user
        self._password = password

    @property
    def access_key_id(self):
        return self._access_key_id

    @property
    def access_key_secret(self):
        return self._access_key_secret

    @property
    def access_key_token(self):
        return self._access_key_token

    @property
    def user(self):
        return self._user

    @property
    def password(self):
        return self._password

    @access_key_token.setter
    def access_key_token(self, v):
        self._access_key_token = v


class _Request(object):

    def __init__(self, container_name):
        self._container_name = container_name

    @property
    def container_name(self):
        """容器(Bucket)名称"""
        return self._container_name

    @container_name.setter
    def container_name(self, v):
        """容器(Bucket)名称"""
        self._container_name = v


class StreamTransferRequest(_Request):

    def __init__(self, container_name, stream, object_name):
        super(StreamTransferRequest, self).__init__(container_name)
        self._stream = stream
        self._object_name = object_name
        self._progress_callback = None
        self._checksum_type = None
        self._data_size = None

    @property
    def stream(self):
        """流式的数据，文件流、网络流、字符串等"""
        return self._stream

    @property
    def object_name(self):
        """对象名称"""
        return self._object_name

    @object_name.setter
    def object_name(self, v):
        self._object_name = v

    @property
    def checksum_type(self):
        """校验码类型[ md5|crc ]  str """
        return self._checksum_type

    @checksum_type.setter
    def checksum_type(self, v):
        self._checksum_type = v

    @property
    def progress_callback(self):
        """进度回调"""
        return self._progress_callback

    @progress_callback.setter
    def progress_callback(self, v):
        self._progress_callback = v

    @property
    def data_size(self):
        """预计的流大小，用于process_callback回调，当不传入时会尝试通过data的属性计算"""
        return self._data_size

    @data_size.setter
    def data_size(self, v):
        self._data_size = v


class FileTransferRequest(_Request):
    def __init__(self, container_name, file_path, object_name):
        super(FileTransferRequest, self).__init__(container_name)

        self._object_name = object_name
        self._file_path = file_path
        self._checksum_type = None
        self._threads_count = 1
        self._progress_callback = None

    @property
    def object_name(self):
        return self._object_name

    @property
    def file_path(self):
        """本地文件路径 str """
        return self._file_path

    @property
    def checksum_type(self):
        """设置校验码类型[ md5|crc ]  str """
        return self._checksum_type

    @property
    def threads_count(self):
        """线程数  int"""
        return self._threads_count

    @property
    def progress_callback(self):
        """进度回调"""
        return self._progress_callback

    @object_name.setter
    def object_name(self, v):
        self._object_name = v

    @file_path.setter
    def file_path(self, v):
        self._file_path = v

    @threads_count.setter
    def threads_count(self, v):
        self._threads_count = v

    @checksum_type.setter
    def checksum_type(self, v):
        self._checksum_type = v

    @progress_callback.setter
    def progress_callback(self, v):
        self._progress_callback = v


class OSSUploadRequest(FileTransferRequest):
    """
    OSS 上传请求，将本地文件上传到oss服务中
    """

    def __init__(self, container_name, file_path, object_name):
        super(OSSUploadRequest, self).__init__(container_name, file_path, object_name)

        self._headers = None
        self._part_size = None

    @property
    def headers(self):
        """可选参数，object的header信息 dict"""
        return self._headers

    @property
    def part_size(self):
        """分片传输时的每一批的大小"""
        return self._part_size

    @headers.setter
    def headers(self, v):
        self._headers = v

    @part_size.setter
    def part_size(self, v):
        self._part_size = v


class FTPUploadRequest(FileTransferRequest):
    """ FTP上传请求，将本地指定文件上传到FTP中

            * object_name: 不包含container目录

            * container_name: ftp中作为container的目录

            * object_name: 不包含container_name的ftp文件路径.

            * checksum_type: checksum type

            * threads_count: FTP无法并发上传文件，threads_count参数无影响

            * buffer_size: 调整ftp通信中传输的数据块大小

        """

    def __init__(self, container_name, file_path, object_name):
        super(FTPUploadRequest, self).__init__(container_name, file_path, object_name)
        self._buffer_size = None

    @property
    def buffer_size(self):
        """ftp上传中传输的数据块大小 int"""
        return self._buffer_size

    @buffer_size.setter
    def buffer_size(self, v):
        self._buffer_size = v


class OSSDownloadRequest(FileTransferRequest):
    """
        OSS下载请求，从存储服务中下载到本地指定目录
    """
    pass


class FTPDownloadRequest(FileTransferRequest):

    """ FTP 下载请求，从ftp中下载文件到本地目录

        * object_name: 不包含container目录

        * container_name: 指定的作为容器的ftp目录

        * file_path: 文件的目标路径.

        """

    def __init__(self, container_name, file_path, object_name):
        super(FTPDownloadRequest, self).__init__(container_name, file_path, object_name)
        self._buffer_size = None

    @property
    def buffer_size(self):
        """ftp下载中传输的数据块大小 int"""
        return self._buffer_size

    @buffer_size.setter
    def buffer_size(self, v):
        self._buffer_size = v


class FTPStreamUploadRequest(StreamTransferRequest):
    """FTP 流式上传请求，从本地上传一个流到FTP服务中"""
    pass


class OSSStreamUploadRequest(StreamTransferRequest):
    """ OSS流式上传请求 """
    def __init__(self, container_name, stream, object_name):
        super(OSSStreamUploadRequest, self).__init__(container_name, stream, object_name)
        self._multi = None

    @property
    def multi(self):
        """是否分片上传, 文件流较大时，可以使用multi参数通知driver使用分片上传 bool"""
        return self._multi

    @multi.setter
    def multi(self, v):
        self._multi = v


class AppendRequest(StreamTransferRequest):
    """ 追加上传请求

        关于追加上传position参数
        * 当Position值为0时，如果没有同名(Appendable) Object，或者同名(Appendable) Object长度为0，该请求成功；
        其他情况均视为Position和Object长度不匹配的情形。

        * 首次追加操作的position必须为0，后续追加操作的position是Object的当前长度。
          例如，第一次Append Object请求指定position值为0，content-length是65536；
          那么，第二次Append Object需要指定position为65536。

    """

    def __init__(self, container_name, stream, object_name, position):
        super(AppendRequest, self).__init__(container_name, stream, object_name)
        self._position = position

    @property
    def position(self):
        """ 续传文件的坐标，必须与存储服务中存在的文件大小相匹配， 新的可追加文件Position为0 int"""
        return self._position

    @position.setter
    def position(self, v):
        self._position = v


class CopyRequest(_Request):
    """从当前Container中拷贝一个Object到当指定的Container

    container_name: 原文件所在的容器
    """

    def __init__(self, container_name, object_name, dst_object_name, dst_container_name=None):
        super(CopyRequest, self).__init__(container_name)
        self._object_name = object_name
        self._dst_container_name = dst_container_name
        self._dst_object_name = dst_object_name

    @property
    def object_name(self):
        """原文件名 str"""
        return self._object_name

    @object_name.setter
    def object_name(self, v):
        self._object_name = v

    @property
    def dst_container_name(self):
        """目的容器，默认为当前容器 str"""
        return self._dst_container_name

    @dst_container_name.setter
    def dst_container_name(self, v):
        self._dst_container_name = v

    @property
    def dst_object_name(self):
        """目的文件名 str"""
        return self._dst_object_name

    @dst_object_name.setter
    def dst_object_name(self, v):
        self._dst_object_name = v


class OSSCopyRequest(CopyRequest):
    def __init__(self, container_name, object_name, dst_object_name, dst_container_name=None):
        super(OSSCopyRequest, self).__init__(container_name, object_name, dst_object_name, dst_container_name)
        self._headers = None
        self._multi_threshold = None

    @property
    def headers(self):
        """object header"""
        return self._headers

    @headers.setter
    def headers(self, v):
        """
        :type v: dict
        :return:
        """
        self._headers = v

    @property
    def multi_threshold(self):
        """设置使用分片复制的阈值"""
        return self._multi_threshold

    @multi_threshold.setter
    def multi_threshold(self, v):
        """

        :type v: int
        :return:
        """
        self._multi_threshold = v


class StreamDownloadRequest(_Request):
    """ 流式下载请求
    """

    def __init__(self, container_name, object_name):
        super(StreamDownloadRequest, self).__init__(container_name)
        self._object_name = object_name
        self._buffer_size = None
        self._progress_callback = None

    @property
    def object_name(self):
        """对象名称"""
        return self._object_name

    @object_name.setter
    def object_name(self, v):
        self._object_name = v

    @property
    def progress_callback(self):
        """进度回调"""
        return self._progress_callback

    @progress_callback.setter
    def progress_callback(self, v):
        self._progress_callback = v

    @property
    def buffer_size(self):
        """流式下载中，从流中取出的每个个数据块大小 int"""
        return self._buffer_size

    @buffer_size.setter
    def buffer_size(self, v):
        self._buffer_size = v


class FTPStreamDownloadRequest(StreamDownloadRequest):
    """ FTP 流式下载请求，获取一个可以持续取出数据的流 """
    pass


class OSSStreamDownloadRequest(StreamDownloadRequest):
    """ OSS 流式下载请求，获取一个可以持续取出数据的流 """
    def __init__(self, container_name, object_name, headers=None, byte_range=None):
        super(OSSStreamDownloadRequest, self).__init__(container_name, object_name)
        self._headers = headers
        self._byte_range = byte_range

    @property
    def headers(self):
        """可选参数，object的header信息 dict"""
        return self._headers

    @headers.setter
    def headers(self, v):
        self._headers = v

    @property
    def byte_range(self):
        """下载范围: (left, right)
            - byte_range 为 (0, 99)  -> 'bytes=0-99'，表示读取前100个字节
            - byte_range 为 (None, 99) -> 'bytes=-99'，表示读取最后99个字节
            - byte_range 为 (100, None) -> 'bytes=100-'，表示读取第101个字节到文件结尾的部分（包含第101个字节）
            - left 和 right均为整数时，0 <= left < right
        """
        return self._byte_range

    @byte_range.setter
    def byte_range(self, v):
        if not isinstance(v, tuple):
            raise TypeError('byte range must be `tuple` type')
        self._byte_range = v


class OperationResult(object):
    def __init__(self, **kwargs):
        self.extra = {}
        if kwargs:
            self.extra.update(**kwargs)

    def __str__(self):
        return "extra: {}".format(self.extra)


class DownloadResult(OperationResult):
    def __init__(self, server_checksum=None, client_checksum=None, checksum_type=None, **kwargs):
        super(DownloadResult, self).__init__(**kwargs)
        self.__server_checksum = server_checksum
        self.__client_checksum = client_checksum
        self.__checksum_type = checksum_type

    @property
    def server_checksum(self):
        return self.__server_checksum

    @property
    def client_checksum(self):
        return self.__client_checksum

    @property
    def checksum_type(self):
        return self.__checksum_type

    def __str__(self):
        info = 'server checksum: {server_checksum}, client checksum: {client_checksum}, ' \
               'checksum type: {checksum_type}. more - {super_str}'
        return info.format(
            server_checksum=self.server_checksum,
            client_checksum=self.client_checksum,
            checksum_type=self.checksum_type,
            super_str=super(DownloadResult, self).__str__()
        )


class StreamDownloadResult(DownloadResult):
    """
        iterable object
    """

    def __init__(self, iterator, length=None, **kwargs):
        super(StreamDownloadResult, self).__init__(**kwargs)
        self.iterator = iterator
        self._length = length
        # 覆盖super.__client_checksum, 通过iterator.
        self.__client_checksum = None

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        try:
            return next(self.iterator)
        except StopIteration:
            if isinstance(self.iterator, io_util.Adapter) and self.iterator.checksum:
                self.__client_checksum = self.iterator.checksum
            raise

    @property
    def client_checksum(self):
        return self.__client_checksum or super(StreamDownloadResult, self).client_checksum

    @property
    def stream(self):
        """将result通过适配器转为file like的对象"""
        return io_util.make_file_like_adapter(self)

    def __len__(self):
        """流的长度"""
        return self._length


class UploadResult(OperationResult):
    def __init__(self, checksum, checksum_type, **kwargs):
        super(UploadResult, self).__init__(**kwargs)
        self.__checksum = checksum
        self.__checksum_type = checksum_type

    def __str__(self):
        info = "checksum: {client_checksum}, type: {checksum_type}, more - {super_info}"
        return info.format(
            client_checksum=self.checksum,
            checksum_type=self.checksum_type,
            super_info=super(UploadResult, self).__str__()
        )

    @property
    def checksum(self):
        return self.__checksum

    @property
    def checksum_type(self):
        return self.__checksum_type


class AppendResult(UploadResult):
    """
    追加上传结果

        * next_position: 下次追加写的偏移 (next_position == object_meta.size)
    """
    def __init__(self, next_position, **kwargs):
        super(AppendResult, self).__init__(**kwargs)
        self.next_position = next_position

    def __str__(self):
        return "next_position: {next}, {super_info}".format(
            next=self.next_position,
            super_info=super(AppendResult, self).__str__()
        )


class RestoreResult(OperationResult):
    def __init__(self, finished, expiry_date=None, **kwargs):
        super(RestoreResult, self).__init__(**kwargs)
        self.finished = finished
        self.expiry_date = expiry_date

    def __str__(self):
        return "restore finished: {finished}, expiry date: {expiry_date}, {super_info}".format(
            finished=self.finished,
            expiry_date=self.expiry_date,
            super_info=super(RestoreResult, self).__str__()
        )


class PartUploadRequest(StreamTransferRequest):
    """
    分片上传请求

    :param stream: 待上传数据。
    """

    def __init__(self, container_name, stream, object_name, part_number, upload_id):

        super(PartUploadRequest, self).__init__(container_name, stream, object_name)
        self._part_num = part_number
        self._upload_id = upload_id

    @property
    def part_number(self):
        """分片号，最小值是1."""
        return self._part_num

    @property
    def upload_id(self):
        """分片上传ID"""
        return self._upload_id


class PartUploadResult(UploadResult):

    def __init__(self, part_info, **kwargs):
        super(PartUploadResult, self).__init__(**kwargs)
        self._part_info = part_info

    @property
    def part_info(self):
        """分片信息.
        :rtype: gddriver.models.PartInfo
        """
        return self._part_info


class PartInfo(object):
    def __init__(self, etag, part_number):
        """

        :type etag: str
        :type part_number: int
        """
        self._etag = etag
        self._part_num = part_number

    @property
    def part_number(self):
        """分片号，最小值是1."""
        return self._part_num

    @property
    def etag(self):
        """分片的ETag，用于合并分块"""
        return self._etag

    @staticmethod
    def sort_parts(parts):
        """为PartInfo列表排序,
        :type parts: list
        """

        def take_part_num(part):
            return part.part_number

        parts.sort(key=take_part_num)
        return parts

    def __str__(self):
        return "Part {}, ETag: {}".format(self.part_number, self.etag)


class CompleteMultipartUploadRequest(_Request):
    def __init__(self, container_name, object_name, upload_id, parts):
        """
        合并分片上传的结果
        :param container_name: bucket/容器名称
        :param object_name: 对象名称
        :param upload_id: 分片上传ID
        :param parts: 所有的分片列表，元素可以是字典或PartInfo
        """
        super(CompleteMultipartUploadRequest, self).__init__(container_name)
        self._object_name = object_name
        self._upload_id = upload_id
        self._parts = self._map_value(parts)

    @property
    def parts(self):
        """list of PartInfo"""
        return self._parts

    @parts.setter
    def parts(self, value):
        """`parts` must be a dict(etag=?, part_number=?) or gddriver.models.PartInfo
        :type value list of dict or list of PartInfo
        """

        self._parts = self._map_value(value)

    @property
    def upload_id(self):
        """分片上传ID"""
        return self._upload_id

    @property
    def object_name(self):
        return self._object_name

    @staticmethod
    def _map_value(args):
        def map_to_parts(v):
            if isinstance(v, PartInfo):
                return v
            elif not isinstance(v, dict):
                raise ValueError("Parameter `parts` must be a dict(etag=?, part_number=?) or gddriver.models.PartInfo")
            return PartInfo(**v)
        return list(map(map_to_parts, args))



# -*- coding: utf-8 -*-

"""
FTPStorageDriver：
    * 列出Bucket中的Objects
    * 获取Object的meta信息
    * 文件上传
    * 流式上传
    * 下载到本地文件
    * 流式下载
    * 复制
    * 删除
支持在request中指定``checksum_type``，在传输完成后返回本地计算的校验码结果

1. 关于container name:
                FTP指定一个目录作为容器目录，
                driver的每次操作都会`change work directory`到container目录，
                container 默认为根目录 /

2. 关于object name:
                FTP的object name为container目录下的相对路径，不能以"/"开头：
                container_name: root, ftp路径： /root/a/b/c， object_name: a/c/c

3. 关于list container:
                * prefix为object的上级目录等：
                   例如： a/b/c  -> prefix可以为 a, a/b, a/b/c等，"folder_a/folder_b" 不能写成 "folder_a/fol"
                * 返回的next_marker永远为None，因为ftp server不支持list部分文件，如果想限制返
                回的数量，可以通过iterator_container。
                * iterator和list都是按层遍历，不会递归遍历所有的文件

4. 关于append_file:
                ftp存储中任何已存在或不存在文件，均可以使用append方法追加

5. ftp 返回码： https://en.wikipedia.org/wiki/List_of_FTP_server_return_codes

"""

import ftplib
import os

import functools
import socket
import uuid

import gddriver.base as base
import gddriver.config as config
import gddriver.errors as gderrors
import gddriver.models as models
import gddriver.utils.io as ioutil
import gddriver.utils.time as timeutil

DEFAULT_CONNECT_TIMEOUT = 30  # 秒, 默认超时时间
DEFAULT_BUFFER_SIZE = 1 << 20  # 1 MB  默认的数据块缓冲大小
DEFAULT_MULTIPART_CHECKSUM_TYPE = 'crc'  # 分片上传时的数据校验方式

_DEFAULT_LOGGER_NAME = __name__


_default_logger = config.get_logger(_DEFAULT_LOGGER_NAME)


def exception_handler(func):

    @functools.wraps(func)
    def ftp_wrapper(*args, **kwargs):
        logger = config.get_logger(_DEFAULT_LOGGER_NAME)
        try:
            return func(*args, **kwargs)
        except (ftplib.error_perm, ftplib.error_reply, ftplib.error_temp) as e:

            logger.debug("ftp driver operation exception occurred.",
                         exc_info=True)

            status_code, msg = _error_parse(e)
            # 530 Not logged in.
            if int(status_code / 10) == 53:
                raise gderrors.NotLoggedIn(
                    message=msg,
                    status=status_code
                )
            elif status_code == 550:
                raise gderrors.NotFound(message=msg)
            raise gderrors.DriverServerException(msg, status_code)
        except socket.timeout as e:
            logger.debug("ftp socket timeout exception occurred.",
                         exc_info=True)
            raise gderrors.RequestTimeout(str(e))

    return ftp_wrapper


def _ensure_cwd(client, path):
    resp = client.cwd(path)
    if not str(resp).startswith("250"):
        # 226 Transfer complete.
        _default_logger.debug("CWD: not response: %s (%s)", resp, client)
        try:
            resp = client.getline()
            _default_logger.debug("CWD: finally response: %s (%s)", resp, client)
        except Exception as e:
            _default_logger.warning("CWD: failed to get the last line of client %s: %s", client, e)


class FTPConnection(base.Connection):
    """
    FTPConnection 的构造方法中，user对应了基类的access_id, password对应了基类的access_key
    在参数上修改名称是为了便于理解，在内部使用仍然为access_id和access_key
    """

    def __init__(self, host, port, credential, ssl=False, anonymous=False, pasv=False, timeout=None):
        super(FTPConnection, self).__init__(
            host=host,
            port=port,
            credential=credential
        )
        self.ssl = ssl
        self.anonymous = anonymous
        self.pasv = pasv
        self.client = None
        self.timeout = timeout or DEFAULT_CONNECT_TIMEOUT
        self.logger = config.get_logger(_DEFAULT_LOGGER_NAME)

    def clone(self):
        """
            复制一个Connection，避免使用同一个session

              ftp的连接不能够复用，当host、port、用户名密码不变时，不需要重新创建Connection对象，
          直接保留原来的信息复制一个实例
        :return:
        :rtype: :class:`FTPConnection`
        """
        return FTPConnection(
            host=self.host,
            port=self.port,
            credential=self.credential,
            ssl=self.ssl,
            anonymous=self.anonymous,
            pasv=self.pasv
        )

    @exception_handler
    def get_client(self, container_name):
        """
        获取ftp客户端(FTPClient)
        :param container_name: 对于FTP，可指定一个目录作为容器，每次get_client时会进入容器目录
        :return: 如过存在ftp连接未关闭，则会继续复用这个连接
        :rtype: :class:`ftplib.FTP`
        """
        container_root = _get_container_root(container_name)
        if self.client:
            try:
                _ensure_cwd(self.client, container_root)
                return self.client
            except ftplib.error_temp as e:
                self.logger.debug(e, exc_info=True)
                # 初次出错，可能连接不可用，忽略问题，尝试重新建立连接
                self.client = None

        user, passwd = (None, None) if self.anonymous else (self.credential.user, self.credential.password)

        if self.ssl:
            ftp_client = self._ssl_connect(user=user, password=passwd)
        else:
            ftp_client = self._normal_connect(user=user, password=passwd)

        try:
            _ensure_cwd(ftp_client, container_root)
            self.client = ftp_client
            return self.client
        except ftplib.error_perm as e:
            code, msg = _error_parse(e)
            if "change directory" in msg:
                self.logger.warning('ftp change directory failed', exc_info=True)
                raise gderrors.NoSuchContainer(container_name)
            else:
                raise

    def _ssl_connect(self, user, password):
        #  ACCT (ACCOUNT)此命令的参数部分使用一个Telnet字符串来指明用户的账户
        ftp = ftplib.FTP_TLS(timeout=self.timeout)
        ftp.set_pasv(self.pasv)
        ftp.connect(self.host, int(self.port))
        # 如果不使用send_cmd的形式登录，抛出异常ftplib.error_perm: 530 Please login with USER and PASS.
        ftp.sendcmd('USER {}'.format(user))
        ftp.sendcmd('PASS {}'.format(password))
        return ftp

    def _normal_connect(self, user, password):
        ftp = ftplib.FTP(timeout=self.timeout)
        ftp.set_pasv(self.pasv)
        ftp.connect(self.host, int(self.port))
        ftp.login(user, password)
        return ftp

    def close(self):
        if self.client:
            try:
                self.client.quit()
            except ftplib.all_errors:
                self.client.close()
            self.client = None


class FTPStorageDriver(base.StorageDriver):
    connection_class = FTPConnection

    def __init__(self):
        super(FTPStorageDriver, self).__init__()

    def list_container_objects(self, connection, container_name, prefix=None, **kwargs):
        """

        返回指定容器内的对象列表

        :param connection:
        :type  connection: :class:`FTPConnection`

        :param container_name: FTP指定一个目录作为容器目录
                                driver的每次操作都会`change work directory`到container目录，
                                container 默认为根目录 /
        :type container_name: :class:`str`

        :param prefix: prefix为object的上级目录等：
                   例如： a/b/c  -> prefix可以为 a, a/b, a/b/c等，"folder_a/folder_b" 不能写成 "folder_a/fol"
        :type  prefix: :class:`str`

        :return: A list of Object instances, next_marker.
        :rtype: ``list`` of :class:`Object` | ``str``

        """
        generator = self.iterate_container_objects(connection, container_name, prefix)
        return [x for x in generator], None

    @exception_handler
    def iterate_container_objects(self, connection, container_name, prefix=None):
        """
        返回指定容器中文件的迭代器

        :param connection:
        :type  connection: :class:`FTPConnection`

        :param container_name: 指定的容器名称.
        :type container_name: :class:`str`

        :param prefix:  5965f2100dd5420091262c06/000000000
        :type  prefix: :class:`str`

        :return: 对象的迭代器
        :rtype: collections.Iterable[Object]
        """
        ftp_client = connection.get_client(container_name)
        container_root = _get_container_root(container_name)
        if prefix:
            _change_work_dir(prefix, ftp_client)
            full_ftp_parent = _normal_full_path(container_root, prefix)
        else:
            full_ftp_parent = container_root
        name_list = ftp_client.nlst()
        for item in name_list:
            is_file, size = _get_file_type_and_size(
                ftp_client=ftp_client,
                name=item,
                sure_exists=True,
                object_name=_join_full_name(prefix, item)
            )
            extra = {
                'parent': full_ftp_parent,
                'short_name': item,
                'is_file': is_file
            }

            prefix = prefix or ''
            obj = models.Object(
                name=_join_full_name(prefix, item),
                size=size,
                **extra
            )
            yield obj

    @exception_handler
    def get_object_meta(self, connection, object_name, container_name):
        """

        :param connection: FTPConnection
        :type connection: :class:`FTPConnection`
        :param object_name: 不包含container目录
        :param container_name: 指定的作为容器的ftp目录
        :return:
        """

        ftp = connection.get_client(container_name)
        container_root = _get_container_root(container_name)
        parent, name = _decomposition_path(object_name)
        full_ftp_path_parent = _normal_full_path(container_root, parent)
        try:
            _change_work_dir(full_ftp_path_parent, ftp)
        except gderrors.NotFound:
            raise gderrors.NoSuchObject(object_name)

        is_file, size = _get_file_type_and_size(
            ftp_client=ftp,
            name=name,
            sure_exists=False,
            object_name=object_name
        )

        extra = {'parent': full_ftp_path_parent, 'is_file': is_file, 'short_name': name}
        return models.Object(
            name=object_name,
            size=size,
            **extra
        )

    @exception_handler
    def download_object_as_stream(self, connection, request):
        """
        FTP流式下载自带了下载的缓冲，默认为64MB。在用户没有及时取出ftp中获取到的数据使用时，下载的数据会存到缓冲区中，
        直到缓冲区满发生阻塞

        :param connection: FTPConnection
        :type connection: :class:`FTPConnection`

        :param request: 流式下载请求
        :type  request: gddriver.models.StreamDownloadRequest

        :return: A generator of file chunk
        :rtype: gddriver.models.StreamDownloadResult
        """
        logger = self.logger
        buffer_size = request.buffer_size
        container_name = request.container_name
        object_name = request.object_name
        progress_callback = request.progress_callback

        obj = self.get_object_meta(connection, object_name, container_name)
        data_size = obj.size

        final_buffer_size = buffer_size
        if not buffer_size or buffer_size < 0:
            final_buffer_size = DEFAULT_BUFFER_SIZE

        stopwatch = timeutil.Stopwatch()
        msg = "download object as stream, container={}, object-key={}, object-size={}, " \
              "buffer-size={}".format(container_name, object_name, obj.size, final_buffer_size)
        logger.debug(msg)
        stopwatch.start()

        def generator():
            ftp = connection.get_client(container_name)
            # 通过RETR命令建立下载文件的连接，建立完成后从该连接中获取数据
            conn = ftp.transfercmd('RETR {}'.format(object_name))
            while True:
                try:
                    data = conn.recv(final_buffer_size)
                except Exception:
                    conn.close()
                    raise
                if not data:
                    break
                yield data
            logger.info("%s, elapsed=%s, download finished", msg, stopwatch.elapsed())
            conn.close()

        stream = generator()
        progress_wrapped_stream = ioutil.make_progress_adapter(
            data=stream,
            progress_callback=progress_callback,
            data_size=data_size
        )
        return models.StreamDownloadResult(progress_wrapped_stream, length=data_size)

    @exception_handler
    def download_file(self, connection, request):
        """
        下载文件，对流式下载的封装

        :param connection: FTPConnection
        :type connection: :class:`FTPConnection`

        :param request: 流式下载请求
        :type  request: gddriver.models.FTPDownloadRequest

        """
        logger = self.logger

        container_name = request.container_name
        object_name = request.object_name
        file_path = request.file_path
        progress_callback = request.progress_callback

        if hasattr(request, 'buffer_size'):
            buffer_size = request.buffer_size or DEFAULT_BUFFER_SIZE
        else:
            buffer_size = DEFAULT_BUFFER_SIZE

        stream_download_request = models.FTPStreamDownloadRequest(
            container_name=container_name,
            object_name=object_name
        )
        stream_download_request.container_name = container_name
        stream_download_request.object_name = object_name
        stream_download_request.progress_callback = progress_callback
        stream_download_request.buffer_size = buffer_size

        generator = self.download_object_as_stream(
            connection=connection,
            request=stream_download_request
        )

        msg = "download to file, bucket={}, object-key={}, " \
              "dst-path={}, buffer-size={}".format(container_name, object_name, file_path, buffer_size)

        logger.debug(msg)
        with open(file_path, 'wb') as f:
            for data in generator:
                f.write(data)

    @exception_handler
    def upload_object_via_stream(self, connection, request):
        """
        流式上传，可以上传网络流、文件流、字符串、bytes等可迭代的对象。

        :param connection: FTPConnection
        :type connection: :class:`FTPConnection`

        :param request:
        :type  request: `gddriver.models.FTPStreamUploadRequest`

        :rtype: ``gddriver.models.UploadResult``
        """
        logger = self.logger

        container_name = request.container_name
        object_name = request.object_name
        stream = request.stream
        progress_callback = request.progress_callback
        checksum_type = request.checksum_type
        data_size = request.data_size

        logger.debug("Upload object %s via stream", object_name)
        upload_chunk_size = int(DEFAULT_BUFFER_SIZE / 2)
        checksum, checksum_type = self.__upload_via_stream(
            cmd='STOR',
            connection=connection,
            object_name=object_name,
            stream=stream,
            container_name=container_name,
            checksum_type=checksum_type,
            chunk_size=upload_chunk_size,
            progress_callback=progress_callback,
            data_size=data_size,
            logger=logger
        )

        return models.UploadResult(
            checksum_type=checksum_type,
            checksum=checksum
        )

    @exception_handler
    def upload_file(self, connection, request):
        """
        上传文件到FTP中

        :param connection:
        :type  connection: :class:`FTPConnection`

        :param request:
        :type  request: `gddriver.models.FTPUploadRequest`

        :rtype: ``UploadResult``
        """
        logger = self.logger

        container_name = request.container_name
        file_path = request.file_path
        object_name = request.object_name
        progress_callback = request.progress_callback
        checksum_type = request.checksum_type

        if hasattr(request, 'buffer_size'):
            buffer_size = request.buffer_size or DEFAULT_BUFFER_SIZE
        else:
            buffer_size = DEFAULT_BUFFER_SIZE

        logger.debug("Upload file %s to %s/%s", file_path, container_name, object_name)

        with open(file_path, 'rb') as f:
            checksum, checksum_type = self.__upload_via_stream(
                cmd='STOR',
                connection=connection,
                object_name=object_name,
                stream=f,
                container_name=container_name,
                checksum_type=checksum_type,
                progress_callback=progress_callback,
                chunk_size=buffer_size,
                logger=logger
            )
            return models.UploadResult(
                checksum=checksum,
                checksum_type=checksum_type
            )

    @exception_handler
    def append_object(self, connection, request):
        """
        追加上传文件，FTP对追加文件没有限制，已存在的任何文件都可以以二进制的形式追加上传，但是只能从文件
        尾部开始追加

        :param connection: FTPConnection
        :type connection: :class:`FTPConnection`

        :param request:
        :type  request: `gddriver.models.AppendRequest`

        :rtype: ``AppendResult``
        """
        logger = self.logger

        container_name = request.container_name
        object_name = request.object_name
        stream = request.stream
        progress_callback = request.progress_callback
        position = request.position
        checksum_type = request.checksum_type
        data_size = request.data_size

        logger.debug("Append object %s via stream", object_name)

        try:
            obj = self.get_object_meta(connection, object_name, container_name)
        except gderrors.NoSuchObject:
            obj = None
            logger.debug("Append object %s via stream, no such object", object_name)

        if obj and position != obj.size:
            raise gderrors.AppendPositionConflict(position)

        checksum, checksum_type = self.__upload_via_stream(
            cmd='APPE',
            connection=connection,
            object_name=object_name,
            stream=stream,
            container_name=container_name,
            checksum_type=checksum_type,
            progress_callback=progress_callback,
            data_size=data_size,
            logger=logger
        )

        obj = self.get_object_meta(connection, object_name, container_name)
        return models.AppendResult(
            next_position=obj.size,
            checksum_type=checksum_type,
            checksum=checksum
        )

    @exception_handler
    def delete_object(self, connection, object_name, container_name=None):
        """
        删除文件，文件不存在时会抛出异常
        :param connection: FTPConnection
        :type connection: :class:`FTPConnection`

        :param object_name: 不包含container名称
        :type object_name: :class:`str`

        :param container_name: 指定的作为容器的ftp目录
        :type container_name: :class:`str`

        """

        ftp = connection.get_client(container_name)
        ftp.delete(object_name)

    @exception_handler
    def copy_object(self, connection, request):
        """

        :param connection:
        :type  connection: :class:`FTPConnection`
        :param request: 对象复制请求
        :type  request: :class:``gddriver.models.CopyRequest``
        :return:
        """
        logger = self.logger

        src_container_name = request.container_name
        src_object_name = request.object_name
        # 默认复制到当前容器中
        dst_container_name = request.dst_container_name or src_container_name
        dst_object_name = request.dst_object_name

        msg = "copy object from {}/{} to {}/{}".format(src_container_name, src_object_name,
                                                       dst_container_name, dst_object_name)
        logger.debug(msg)
        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        stream_download_request = models.FTPStreamDownloadRequest(
            container_name=src_container_name,
            object_name=src_object_name
        )
        stream_download_request.buffer_size = DEFAULT_BUFFER_SIZE
        stream = self.download_object_as_stream(
            connection=connection,
            request=stream_download_request
        )

        stream_upload_request = models.FTPStreamUploadRequest(
            container_name=dst_container_name,
            stream=stream,
            object_name=dst_object_name
        )

        stream_upload_request.checksum_type = 'md5'
        # 必须复制一个连接，否则会阻塞
        copy_connection = connection.clone()
        try:
            res = self.upload_object_via_stream(
                connection=copy_connection,
                request=stream_upload_request
            )

            logger.debug("%s, res: %s, elapsed=%s, copy finished", msg, res, stopwatch.elapsed())

        finally:
            copy_connection.close()

    def restore_object(self, connection, src_container_name, src_object_name, **kwargs):
        raise NotImplementedError

    def archive_object(self, connection, src_container_name, src_object_name, archive_container_name,
                       archive_object_name,
                       **kwargs):
        raise NotImplementedError

    @exception_handler
    def move_object(self, connection, src_object_name, dst_object_name, container_name=None, dst_container_name=None):
        """
        重命名： src_container/src_obj -> dst_container/dst_obj
        NOTE: raise 250异常时，多是因为复用ftp_client造成，使用connection.clone复制一个连接可避免该问题

        :param connection:
        :type  connection: :class:`Connection`

        :param src_object_name: 原始对象名
        :type src_object_name: `str`

        :param dst_object_name: 目标对象名称
        :type dst_object_name: `str`

        :param container_name: 原文件所在的容器 （默认为ftp存储的根目录 "/" ）
        :type  container_name: :class:`str`

        :param dst_container_name: 目标容器（默认原文件所在容器）
        :type dst_container_name: `str`

        """
        msg = "move object from {}/{} to {}/{}".format(container_name, src_object_name,
                                                       dst_container_name, dst_object_name)
        self.logger.debug(msg)
        ftp_client = connection.get_client(container_name)

        def build_prefix(name):
            return '/' + name if name else '/'

        prefix_src = build_prefix(container_name)
        prefix_dst = build_prefix(dst_container_name) if dst_container_name else prefix_src
        src_path = '/'.join([prefix_src, src_object_name])
        dst_path = '/'.join([prefix_dst, dst_object_name])
        res = ftp_client.rename(src_path, dst_path)
        self.logger.info(res)

    @exception_handler
    def init_multipart_upload(self, connection, object_name, container_name=None, overwrite=False):
        """
        初始化分片上传 （ftp服务中创建 object_name.multi-uuid 目录用于分片上传缓存）
        :type connection: FTPConnection
        :type object_name: str
        :type container_name: str
        :param overwrite: 覆盖原文件
        :rtype: str
        """

        upload_id = "multi-{}".format(uuid.uuid1())
        if overwrite:
            try:
                self.delete_object(
                    connection=connection,
                    object_name=object_name,
                    container_name=container_name
                )
            except gderrors.NotFound as e:
                self.logger.debug("file not found while initializing multi-upload and overwrite %s: %s", object_name, e)
        multi_tmp_dir = self.__get_multi_tmp_dir_with_check(
            connection=connection,
            container_name=container_name,
            object_name=object_name,
            upload_id=upload_id,
            init=True
        )

        # create tmp dir
        self.check_or_make_dir(
            connection=connection,
            container_name=container_name,
            path=multi_tmp_dir
        )
        self.logger.debug("multipart upload tmp directory %s created", multi_tmp_dir)

        return upload_id

    @exception_handler
    def upload_part(self, connection, request):
        """
        分片上传，上传时默认计算分片的crc校验码并作为ETag返回（合并分片时通过ETag中的校验码对分片进行校验）

        :type connection: FTPConnection
        :type request: gddriver.models.PartUploadRequest
        :return:
        """
        multi_tmp_dir = self.__get_multi_tmp_dir_with_check(
            connection=connection,
            container_name=request.container_name,
            object_name=request.object_name,
            upload_id=request.upload_id
        )

        part_name = '/'.join([multi_tmp_dir, str(request.part_number)])
        stream_upload_request = models.FTPStreamUploadRequest(
            container_name=request.container_name,
            stream=request.stream,
            object_name=part_name
        )
        stream_upload_request.progress_callback = request.progress_callback
        stream_upload_request.checksum_type = DEFAULT_MULTIPART_CHECKSUM_TYPE
        result = self.upload_object_via_stream(
            connection=connection,
            request=stream_upload_request
        )

        # etag统一为string
        part_etag = self.__checksum_to_etag(result.checksum)
        return models.PartUploadResult(
            checksum_type=result.checksum_type,
            checksum=result.checksum,
            part_info=models.PartInfo(etag=part_etag, part_number=request.part_number)
        )

    @exception_handler
    def complete_multipart_upload(self, connection, request):
        """ 合并所有分块，当目标文件已存在时或分块文件不存在时抛出异常
        throws:
            FileAlreadyExists  object_name对应的文件已存在
            BadRequest  upload_id对应的临时目录不存在或分片参数有问题

        :type connection: FTPConnection
        :type request: gddriver.models.CompleteMultipartUploadRequest
        """

        multi_tmp_dir = self.__get_multi_tmp_dir_with_check(
            connection=connection,
            container_name=request.container_name,
            object_name=request.object_name,
            upload_id=request.upload_id
        )

        parts = models.PartInfo.sort_parts(request.parts)
        tmp_name = '/'.join([multi_tmp_dir, "tmp"])

        def _merge_parts():
            """按照分片的顺序将各个分片文件append到同一个文件中"""
            position = 0
            for part in parts:
                part_name = '/'.join([multi_tmp_dir, str(part.part_number)])
                stream_download_request = models.FTPStreamDownloadRequest(
                    container_name=request.container_name,
                    object_name=part_name
                )
                stream_download_request.buffer_size = DEFAULT_BUFFER_SIZE

                try:
                    part_stream = self.download_object_as_stream(connection, stream_download_request)
                except gderrors.NotFound as _e:
                    raise gderrors.BadRequest("One or more of the specified parts could not be found: {}".format(_e))

                conn_clone = connection.clone()
                with conn_clone:
                    append_request = models.AppendRequest(
                        container_name=request.container_name,
                        object_name=tmp_name,
                        stream=part_stream,
                        position=position
                    )
                    append_request.checksum_type = DEFAULT_MULTIPART_CHECKSUM_TYPE
                    append_result = self.append_object(
                        connection=conn_clone,
                        request=append_request
                    )
                    if self.__checksum_to_etag(append_result.checksum) != part.etag:
                        raise gderrors.BadRequest(("The specified entity tag not matched "
                                                   "the part's entity tag, request: {}, actually: {}").format(
                            part.etag, append_result.checksum
                        ))
                position += part_stream.__len__()

        stopwatch = timeutil.Stopwatch()
        stopwatch.start()
        self.logger.debug("start complete multipart upload %s, parts=%s, upload_id=%s",
                          request.object_name, len(request.parts), request.upload_id)
        copy_conn = connection.clone()
        success = False
        with copy_conn:
            try:
                # 合并分片到一个临时文件
                _merge_parts()

                # rename tmp -> object_name
                self.move_object(copy_conn, tmp_name, request.object_name, container_name=request.container_name)

                # 删除临时目录
                self.delete_dir(
                    container_name=request.container_name,
                    connection=copy_conn,
                    dir_name=multi_tmp_dir
                )
                success = True
            finally:
                self.logger.debug("finished complete multipart upload %s, parts=%s, upload_id=%s, time-elapsed=%s",
                                  request.object_name, len(request.parts), request.upload_id, stopwatch.elapsed())
                if success:
                    return
                # 失败时删除临时文件
                try:
                    self.delete_object(
                        connection=copy_conn,
                        container_name=request.container_name,
                        object_name=tmp_name
                    )
                except gderrors.NotFound:
                    pass
                except gderrors.DriverServerException as e:
                    self.logger.debug("An exception occurred during delete %s: %s", tmp_name, e)

    def batch_delete_objects(self, connection, object_name_list, container_name=None):
        raise NotImplementedError

    def get_object_sign_url(self, connection, object_name, container_name=None, **kwargs):
        raise NotImplementedError

    def __get_multi_tmp_dir_with_check(self, connection, container_name, object_name, upload_id, init=False):
        """获取分片上传的临时目录，并同时检查upload_id是否有效"""
        try:
            meta = self.get_object_meta(
                container_name=container_name,
                connection=connection,
                object_name=object_name
            )
            if meta:
                raise gderrors.FileAlreadyExists("File {} already exists".format(meta))
        except gderrors.NotFound:
            pass

        multi_tmp_dir = self.__multi_tmp_dir(object_name, upload_id)
        if not init:
            try:
                dir_meta = self.get_object_meta(
                    connection=connection,
                    container_name=container_name,
                    object_name=multi_tmp_dir
                )
                self.logger.debug("Found multipart upload process: {}".format(dir_meta))
            except gderrors.NotFound as e:
                self.logger.warning("Not found such upload_id: %s", e)
                raise gderrors.BadRequest("Not found such upload_id {} with object {}".format(
                    upload_id, object_name))
        return multi_tmp_dir

    @staticmethod
    def __checksum_to_etag(checksum):
        """将crc/md5转为etag，当前默认转为str类型"""
        return str(checksum)

    @staticmethod
    def __multi_tmp_dir(object_name, upload_id):
        """
        format: object_name.multi-uuid

        :param str object_name: file path
        :param str upload_id: multipart upload id
        :rtype: str
        """
        return '.'.join([object_name, upload_id])

    @classmethod
    def __upload_via_stream(cls, cmd, connection, stream, container_name, object_name, checksum_type=None,
                            progress_callback=None, data_size=None, logger=None, chunk_size=None):

        parent, name = _decomposition_path(object_name)
        cls.check_or_make_dir(
            connection=connection,
            container_name=container_name,
            path=parent
        )

        ftp_client = connection.get_client(container_name)
        _change_work_dir(parent, ftp_client)

        progress_wrapped_stream = ioutil.make_progress_adapter(
            data=stream,
            progress_callback=progress_callback,
            data_size=data_size
        )
        checksum_wrapped_stream = ioutil.make_checksum_adapter(
            progress_wrapped_stream,
            checksum_type=checksum_type
        )

        readable_stream = ioutil.make_file_like_adapter(checksum_wrapped_stream)

        # 开始传输
        msg = "upload via stream, container={}, object-key={}, checksum-type={}, " \
              "data-size={}".format(container_name, object_name, checksum_type, data_size)
        logger.debug(msg)
        stopwatch = timeutil.Stopwatch()
        stopwatch.start()

        # stirbinary: ftp 上传操作，其中包括 STOR(存储), APPE(追加) 等命令
        ftp_client.storbinary(
            cmd='{} {}'.format(cmd, name),
            fp=readable_stream,
            blocksize=chunk_size
        )

        logger.info('%s, checksum=%s, elapsed=%s, upload finished',
                    msg, checksum_wrapped_stream.checksum, stopwatch.elapsed())

        return checksum_wrapped_stream.checksum, checksum_wrapped_stream.checksum_type

    @classmethod
    def check_or_make_dir(cls, connection, container_name, path):
        """
        递归创建目录
        :type  connection: :class:`FTPConnection`
        :type  container_name: :class:`str`
        :type  path: :class:`str`
        """
        if not path:
            return
        container_root = _get_container_root(container_name)
        ftp = connection.get_client(container_name)
        try:
            _ensure_cwd(ftp, path)
        except ftplib.error_perm:
            parent, name = _decomposition_path(path)
            cls.check_or_make_dir(connection, container_name, parent)
            full_parent_path = _normal_full_path(container_root, parent)
            _ensure_cwd(ftp, full_parent_path)
            ftp.mkd(name)

    @classmethod
    @exception_handler
    def delete_dir(cls, connection, container_name, dir_name):
        """
        递归删除目录
        :type  connection: :class:`FTPConnection`
        :type container_name: :class:`str`
        :type  dir_name: :class:`str`
        """
        ftp = connection.get_client(container_name)
        name_list = ftp.nlst(dir_name)
        if name_list:
            for name in name_list:
                try:
                    ftp.size(name)
                    ftp.delete(name)
                except ftplib.error_perm:
                    cls.delete_dir(connection, container_name, name)
        ftp.rmd(dir_name)


def _error_parse(err):
    """
    错误实例：
        530	Not logged in.
    :param err:
    :type :class:`ftplib.Error`
    :return:
    """
    msg_str = str(err)
    if len(msg_str) < 3:
        return None, msg_str

    code = msg_str[:3]
    msg = msg_str[3:]
    try:
        code = int(code)
    except ValueError:
        code = None
        msg = msg_str
    return code, msg


def _change_work_dir(directory, ftp_client):
    if directory == '':
        return

    logger = config.get_logger(_DEFAULT_LOGGER_NAME)
    try:
        logger.debug('(%s) try to change directory to %s.', ftp_client, directory)
        _ensure_cwd(ftp_client, directory)
        logger.debug('current directory changed (%s): %s.', ftp_client, directory)
    except ftplib.error_perm as e:
        status, msg = _error_parse(e)
        # 550 Requested action not taken. File unavailable (e.g., file not found, no access).
        if status == 550:
            raise gderrors.NotFound(
                status=550,
                message="directory not found: {}".format(directory)
            )
        raise


def _decomposition_path(path):
    """ /a/b/c/d -> (a/b/c, d)"""

    if not path:
        return '', ''
    path = os.path.normpath(path)
    sp = path.split('/')

    parent = ''
    name = sp[-1]
    if len(sp) > 1:
        parent = os.path.normpath('/'.join(sp[:-1]))
    return parent, name


def _get_container_root(container_name=None):
    """ /container_name/ """
    return '/{}/'.format(container_name) if container_name else '/'


def _normal_full_path(container_root, path):
    """ /container_name/path/
    :param container_root:  /container_name/
    :param path:            path
    :return 拼接的路径，统一以 '/' 结尾
    """
    full_path = container_root + path
    full_path = os.path.normpath(full_path)
    return full_path + '/' if full_path != '/' else full_path


def _join_full_name(prefix, name):
    if not prefix:
        return name
    if not prefix.endswith('/'):
        return '/'.join([prefix, name])
    else:
        return ''.join([prefix, name])


def _get_file_type_and_size(ftp_client, name, sure_exists=True, object_name=None):
    """ (is_file, file_size)"""
    logger = config.get_logger(_DEFAULT_LOGGER_NAME)
    try:
        logger.debug('try to get the size of %s. (object_name: %s)', name, object_name)
        size = ftp_client.size(name)
        if size is None:
            # FTP lib的bug，有时跟在cwd命令后的命令会无任何返回值，尝试再次获取文件大小
            size = ftp_client.size(name)
        if size is None:
            raise gderrors.DriverServerException("get size failed")
        logger.info('get the size of %s finished. (object_name: %s)', name, object_name)
        return True, size
    except ftplib.error_perm as e:
        status, msg = _error_parse(e)
        if status != 550:
            raise
        size = 0
        is_file = False
        if not sure_exists:
            try:
                # 通过切换目录检查目录是否存在
                pwd = ftp_client.pwd()
                _change_work_dir(name, ftp_client)
                _change_work_dir(pwd, ftp_client)
            except gderrors.NotFound:
                raise gderrors.NoSuchObject(object_name)
        return is_file, size



# -*- coding: utf-8 -*-

from __future__ import with_statement

import gddriver.config as config

# 32 MB
DEFAULT_MULTIPART_TRANSFER_THRESHOLD = 1 << 25

__all__ = [
    'Connection',
    'StorageDriver'
]


class Connection(object):
    """
     Storage Server连接的许可信息，Driver执行操作时，通过connection信息完成服务认证
    """

    def __init__(self, host, port, credential, **kwargs):
        self.host = host
        self.port = port
        self.credential = credential

    def get_client(self, container_name):
        """
          获取认证后的连接客户端。
          在没有关闭连接之前，get_client会复用已有的连接
        """
        raise NotImplementedError

    def clone(self):
        """
          复制一个Connection，避免使用同一个session.
        :return:
        :rtype: :class:`Connection`
        """
        raise NotImplementedError

    def close(self):
        """关闭连接"""
        raise NotImplementedError

    def __str__(self):
        return "Connection<host={}, port={}>".format(self.host, self.port)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class StorageDriver(object):
    connection_class = Connection

    @classmethod
    def create_connection_manager(cls, store_host, store_port, credential, **kwargs):
        """
        创建一个新的连接器，不同的 DriverImpl 会自动初始化各自的 ConnectionImpl

        NOTE: 由于设计的原因，connection的定义不够清析，如今 connection == connection_manager == connector
        :param store_host:
        :param store_port:
        :param credential:
        :type  credential: :class:`gddriver.models.Credential`
        :rtype: gddriver.base.Connection
        """
        return cls.connection_class(store_host, store_port, credential, **kwargs)

    def __init__(self):
        self.logger = config.get_logger("gddriver.{}".format(self.__class__.__name__))

    def get_object_meta(self, connection, object_name, container_name):
        """
        返回存储服务中获取的实例信息

        :param connection:
        :type  connection: :class:`Connection`

        :param container_name: Container name.
        :type  container_name: ``str``

        :param object_name: Object name.
        :type  object_name: ``str``

        :return: :class:`Object` instance.
        :rtype: :class:`gddriver.models.Object`
        """
        raise NotImplementedError

    def upload_file(self, connection, request):
        """
        通过路径上传文件

        :param connection:
        :type  connection: :class:`Connection`

        :param request: 文件上传请求
        :type request: :class:`gddriver.models.FileTransferRequest`

        :return: 上传成功的结果
        :rtype: :class:``gddriver.models.UploadResult``
        """

        raise NotImplementedError

    def upload_object_via_stream(self, connection, request):
        """
        流式上传

        :param connection:
        :type  connection: :class:`Connection`

        :param request: 流式上传请求
        :type request: :class:`gddriver.models.StreamTransferRequest`

        :return: 上传成功的结果
        :rtype: :class:``gddriver.models.UploadResult``

        """
        raise NotImplementedError

    def append_object(self, connection, request):
        """
        将一个流追加上传到指定的object中

        :param connection:
        :type  connection: :class:`Connection`

        :param request: 追加上传请求
        :type request: :class:`gddriver.models.AppendRequest`

        :rtype: ``gddriver.models.AppendResult``
        """
        raise NotImplementedError

    def download_file(self, connection, request):
        """
        下载文件到指定的路径

        :param connection:
        :type  connection: :class:`Connection`

        :param request: 文件下载请求
        :type request: :class:`gddriver.models.FileTransferRequest`

        """
        raise NotImplementedError

    def download_object_as_stream(self, connection, request):
        """
        返回一个流的生成器.

        :param connection:
        :type  connection: :class:`Connection`

        :param request: 流式下载请求
        :type request: :class:`gddriver.models.StreamDownloadRequest`

        :return: A generator of file chunk
        :rtype: :class:``gddriver.models.StreamDownloadResult``

        """
        raise NotImplementedError

    def copy_object(self, connection, request):
        """

        从当前Container中拷贝一个Object到当指定的Container


        :param connection:
        :type  connection: :class:`Connection`

        :param request: 原文件所在的容器(bucket)
        :type  request: :class:``gddriver.models.CopyRequest``

        :return:
        """
        raise NotImplementedError

    def archive_object(self, connection, src_container_name, src_object_name,
                       archive_container_name, archive_object_name, **kwargs):
        """

        从当前容器归档一个文件到目标容器


        :param connection:
        :type  connection: :class:`Connection`

        :param src_container_name: 原文件所在的容器
        :type  src_container_name: :class:`str`

        :param src_object_name: 原始对象名
        :type src_object_name: `str`

        :param archive_container_name: 目标容器
        :type archive_container_name: `str`

        :param archive_object_name: 目标对象名称
        :type archive_object_name: `str`

        """
        raise NotImplementedError

    def restore_object(self, connection, src_container_name, src_object_name):
        """
        :param connection:
        :type  connection: :class:`Connection`

        :param src_object_name: 原始对象名
        :type src_object_name: `str`

        :param src_container_name: 原文件所在的容器
        :type  src_container_name: :class:`str`

        :rtype: :class:`gddriver.models.RestoreResult`
        """
        raise NotImplementedError

    def move_object(self, connection, src_object_name, dst_object_name,
                    container_name=None, dst_container_name=None):
        """
        移动/重命名
        since gddriver 0.2.1

        :param connection:
        :type  connection: :class:`Connection`

        :param src_object_name: 原始对象名
        :type src_object_name: `str`

        :param dst_object_name: 目标对象名称
        :type dst_object_name: `str`

        :param container_name: 原文件所在的容器
        :type  container_name: :class:`str`

        :param dst_container_name: 目标容器（默认原始容器）
        :type dst_container_name: `str`

        """
        raise NotImplementedError

    def delete_object(self, connection, object_name, container_name):
        """
        删除对象

        :param container_name:
        :type container_name: `str`

        :param object_name: Object name.
        :type object_name: `str`

        :param connection: Destination container.
        :type connection: :class:`Connection`

        """
        raise NotImplementedError

    def batch_delete_objects(self, connection, object_name_list, container_name):
        """
        批量删除

        :param connection:
        :type  connection: :class:`Connection`

        :param object_name_list: 待删除列表
        :type object_name_list: list of str

        :param container_name: 容器名称
        :type  container_name: :class:`str`

        """
        raise NotImplementedError

    def list_container_objects(self, connection, container_name, prefix=None, **kwargs):
        """
        返回对象的列表

        :param connection:
        :type  connection: :class:`Connection`

        :param container_name: Destination container.
        :type container_name: :class:`str`

        :param prefix: object_name 的前缀
        :type  prefix: :class:`str`

        :return: A list of Object instances, next_marker.
        :rtype: ``list`` of :class:`Object` | ``str``
        """
        raise NotImplementedError

    def get_object_sign_url(self, connection, object_name, container_name=None, **kwargs):
        raise NotImplementedError

    def init_multipart_upload(self, connection, object_name, container_name=None, overwrite=False):
        """
        初始化分片上传

        :param connection:
        :type  connection: :class:`Connection`

        :param object_name:  Object name
        :type object_name: str

        :param container_name: Destination container.
        :type container_name: :class:`str`

        :param overwrite: Overwrite the exist file
        :type overwrite: bool

        :return: upload_id
        :rtype: str
        """
        raise NotImplementedError

    def upload_part(self, connection, request):
        """
        :param connection:
        :type  connection: :class:`Connection`

        :param request:  request object
        :type request: gddriver.models.PartUploadRequest

        :rtype: gddriver.models.PartUploadResult
        """
        raise NotImplementedError

    def complete_multipart_upload(self, connection, request):
        """
        :param connection:
        :type  connection: :class:`Connection`

        :param request:  request object
        :type request: gddriver.models.CompleteMultipartUploadRequest

        :rtype: gddriver.models.UploadResult
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

__author__ = 'Rao Mengnan'

import os

try:
    import setuptools
except ImportError:
    import distutils.core as setuptools


location = os.path.dirname(os.path.abspath(__file__))
about = {}
with open(os.path.join(location, 'gddriver/__about__.py'), 'r') as f:
    exec(f.read(), about)

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = [
    'crcmod',
    'oss2>=2.4.0'
]

test_requirements = ['mock']

packages = setuptools.find_packages(exclude=['tests', 'tests.*'])
setuptools.setup(
    name='gddriver',
    description="The gddriver is an efficient tool for operating the storage service such as oss and ftp.",
    version=about['VERSION'],
    author='GeneDock Contributor',
    author_email="raomengnan@genedock.com",
    maintainer='GeneDock Contributor',
    maintainer_email='raomengnan@genedock.com',
    url='https://genedock.com',
    packages=packages,
    package_data={'': ['LICENSE', 'requirements.txt']},
    license='ASF',
    long_description=long_description,
    long_description_content_type="text/markdown",

    platforms=['Independent'],
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements,
    tests_require=test_requirements,
    test_suite='tests'
)



# -*- coding: utf-8 -*-


import logging


class _LoggerConfig(object):

    _LOGGER_NAME = "gddriver"

    def __init__(self):
        self._logger = None

    @property
    def logger(self):
        return self._logger

    def get_logger(self, logger_name=None):
        if self.logger:
            return self.logger

        return logging.getLogger(logger_name or self._LOGGER_NAME)


_DEFAULT_LOGGER_CONFIG = _LoggerConfig()


def get_logger(name=None):
    return _DEFAULT_LOGGER_CONFIG.get_logger(name)


def set_logger(logger):
    _DEFAULT_LOGGER_CONFIG._logger = logger
    get_logger().debug("logger changed.")



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import time


class Stopwatch(object):

    def __init__(self):
        self.__is_run = False
        self.__start_time = 0
        self.__elapsed_times = 0

    def start(self):
        if self.__is_run:
            return False
        self.__is_run = True
        self.__start_time = time.time()

    def stop(self):
        if not self.__is_run:
            return False
        self.__is_run = False
        self.__elapsed_times += time.time() - self.__start_time

    def elapsed(self):

        if self.__is_run:
            return self.__elapsed_times + time.time() - self.__start_time
        return self.__elapsed_times

    def reset(self):

        self.__elapsed_times = 0
        self.__start_time = 0


def timestamp_to_date(timestamp):
    return datetime.datetime.utcfromtimestamp(timestamp)



# -*- coding: utf-8 -*-

import os

import gddriver
import gddriver.config as config
import gddriver.errors as errors
import gddriver.models as models
import gddriver.utils.io as ioutil

logger = config.get_logger(__name__)
DEFAULT_MULTIPART_TRANSFER_THRESHOLD = 1 << 25  # 32 MB


class _CommonStreamUploadRequest(models.StreamTransferRequest):
    def __init__(self, container_name, stream, object_name):
        super(_CommonStreamUploadRequest, self).__init__(container_name, stream, object_name)
        self._multi = None

    @property
    def multi(self):
        """是否分片上传 bool"""
        return self._multi

    @multi.setter
    def multi(self, v):
        self._multi = v


def upload_object_from_file(connection, request, driver, multipart_transfer_threshold):
    """
    通过路径上传文件，当文件大小超过阈值时，使用分片上传的形式

    文件超过阈值时，由于分片并发上传，校验码的计算可能需要另启一个线程，
    但使用OSS上传并且要求crc64校验码时就不会有此情况，oss会在上传完成时，将crc64返回

    :param driver:  存储服务驱动
    :type  driver: `gddriver.base.StorageDriver`

    :param connection: 连接信息
    :type  connection: `gddriver.base.Connection`

    :param request: 传输请求
    :type  request: `gddriver.models.FileTransferRequest`

    :param multipart_transfer_threshold: 分片传输阈值
    :type  multipart_transfer_threshold: int

    :return:
    """

    final_multipart_threshold = multipart_transfer_threshold or DEFAULT_MULTIPART_TRANSFER_THRESHOLD
    file_path = request.file_path
    object_name = request.object_name

    file_size = os.path.getsize(file_path)
    if file_size < final_multipart_threshold:
        logger.debug("upload %s to %s via stream, data-size=%s", file_path, object_name, file_size)
        with open(file_path, 'rb') as f:
            # 流式上传时，data_size需要用户自定义或者用户的progress_callback有能力处理data_size为空的情况
            stream_request = _CommonStreamUploadRequest(
                container_name=request.container_name,
                stream=f,
                object_name=object_name
            )

            stream_request.checksum_type = request.checksum_type
            stream_request.progress_callback = request.progress_callback
            stream_request.data_size = file_size

            res = driver.upload_object_via_stream(
                request=stream_request,
                connection=connection
            )
            logger.debug("upload %s to %s via stream, data-size=%s, upload successfully",
                         file_path, object_name, file_size)
            return res

    else:
        logger.debug("upload %s to %s by file path, data-size=%s, threads_count=%s",
                     file_path, object_name, file_size, request.threads_count)
        res = driver.upload_file(
            connection=connection,
            request=request
        )
        logger.debug("finished upload %s to %s by file path, data-size=%s, "
                     "threads_count=%s, upload successfully",
                     file_path, object_name, file_size, request.threads_count)
        return res


def download_object_to_file(connection, request, driver,
                            multipart_transfer_threshold, overwrite_existing, delete_on_failure):
    """
        当文件大小超过阈值(threshold)时，会启用多分片并发下载，如果要求计算校验码，则需要额外的时间同步等待校验码计算完成。
    文件未超过阈值时，校验码可以在传输时一并完成计算。

    :param driver:  存储服务驱动
    :type  driver: `gddriver.base.StorageDriver`

    :param connection: 连接信息
    :type  connection: `gddriver.base.Connection`

    :param request: 传输请求
    :type  request: `gddriver.models.FileTransferRequest`

    :param multipart_transfer_threshold: 分片并发下载的阈值（字节）
    :type  multipart_transfer_threshold: ``int``

    :param overwrite_existing: 本地存在时覆盖
    :type  overwrite_existing: ``bool``

    :param delete_on_failure:  下载失败时删除
    :type  delete_on_failure: ``bool``

    :rtype: :class:`gddriver.models.DownloadResult`
    """
    file_path = request.file_path
    object_name = request.object_name
    final_threshold = multipart_transfer_threshold or DEFAULT_MULTIPART_TRANSFER_THRESHOLD
    checksum_type = request.checksum_type

    info = "Download {object_name} to {file_path}, container {container_name}".format(
        object_name=object_name,
        file_path=file_path,
        container_name=request.container_name
    )

    if os.path.isdir(file_path):
        raise errors.PathTypeConflict("%s, local path is a directory".format(info))

    if os.path.exists(file_path) and not overwrite_existing:
        raise errors.FileAlreadyExists("%s, file already exists".format(info))

    object_meta = driver.get_object_meta(
        connection=connection,
        object_name=object_name,
        container_name=request.container_name
    )
    logger.debug("download %s to %s, data-size-threshold: %s, object info: %s",
                 object_name, file_path, final_threshold, object_meta)

    if object_meta.size < final_threshold:
        stream_download_request = models.StreamDownloadRequest(
            container_name=request.container_name,
            object_name=object_name)

        readable_stream = driver.download_object_as_stream(
            request=stream_download_request,
            connection=connection
        )
        stream_adapter = ioutil.make_checksum_adapter(readable_stream, checksum_type)

        with open(file_path, 'wb') as file_:
            size = ioutil.copy_file(stream_adapter, file_)
            if size != object_meta.size and delete_on_failure:
                os.remove(file_path)
                raise errors.UnexpectedDownloadedSize(object_meta.size, size)

        # 部分服务可能会提供server_checksum (比如OSS提供crc_checksum)
        result = models.DownloadResult(
            server_checksum=readable_stream.server_checksum,
            client_checksum=stream_adapter.checksum,
            checksum_type=stream_adapter.checksum_type
        )
    else:
        driver.download_file(
            request=request,
            connection=connection
        )

        checksum_gen = ioutil.create_checksum_yield(
            file_path=file_path,
            checksum_type=checksum_type,
            logger=logger
        )
        checksum = next(checksum_gen)
        result = models.DownloadResult(
            client_checksum=checksum,
            checksum_type=checksum_type
        )
    logger.info("download %s to %s, container %s, download successfully.",
                object_name, file_path, request.container_name)
    return result


def get_object_append_position(connection, driver, object_name, container_name):
    """
    获取对象当前的长度，可用于append_object的position

    :param connection:
    :type  connection: ``gddriver.base.Connection``
    :param driver:
    :type  driver:  ``gddriver.base.StorageDriver``
    :param container_name:
    :type  container_name: ``str``
    :param object_name: Object name.
    :type object_name: ``str``

    :return: 对象长度/坐标
    :rtype: ``int``
    """
    try:
        object_meta = driver.get_object_meta(
            connection=connection,
            object_name=object_name,
            container_name=container_name
        )
        return object_meta.size
    except errors.NoSuchObject:
        return 0


class GenericOperator(object):
    """
    Generic operator: 根据provider初始化相应的driver，并通过host/port和credential等信息初始化connection manager等信息

    为什么创建一个这样的代理：
        1. 存在使用者不想关注如何创建driver、connection等信息（精简操作步骤）
        2. 存在使用者不想单独管理 connection 和 driver (减少参数)
        3. 支持使用with的上下文，在每次使用之后释放资源
    since gddriver 0.2.1
    """
    def __init__(self, provider, credential, host, port=None, **kwargs):
        """

        :type credential: gddriver.models.Credential
        :param provider: oss | ftp : see gddriver.providers.Provider
        :param host: store host
        :param port: store port
        :param kwargs: 附加信息用于初始化Connection，具体可接受何种附加信息可见 OSSConnection / FtpConnection等实现类
        """
        self._driver = gddriver.get_driver(provider=provider)
        self._connector = self._driver.create_connection_manager(
            store_host=host,
            store_port=port,
            credential=credential,
            **kwargs
        )

    def get_object_meta(self, object_name, container_name):
        """
        返回存储服务中获取的实例信息

        :param container_name: Container name.
        :type  container_name: ``str``

        :param object_name: Object name.
        :type  object_name: ``str``

        :return: :class:`Object` instance.
        :rtype: :class:`gddriver.models.Object`
        """
        return self._driver.get_object_meta(
            self._connector,
            object_name=object_name,
            container_name=container_name
        )

    def list_container_objects(self, container_name, prefix=None, **kwargs):
        """
        返回对象的列表

        :param container_name: Destination container.
        :type container_name: :class:`str`

        :param prefix: object_name 的前缀
        :type  prefix: :class:`str`

        :return: A list of Object instances, next_marker.
        :rtype: ``list`` of :class:`Object` | ``str``
        """
        return self._driver.list_container_objects(
            self._connector,
            container_name=container_name, prefix=prefix,
            **kwargs
        )

    def upload_file(self, request):
        """
        通过路径上传文件

        :param request: 文件上传请求
        :type request: :class:`gddriver.models.FileTransferRequest`

        :return: 上传成功的结果
        :rtype: :class:``gddriver.models.UploadResult``
        """
        return self._driver.upload_file(
            self._connector,
            request=request
        )

    def upload_object_via_stream(self, request):
        """
        流式上传

        :param request: 流式上传请求
        :type request: :class:`gddriver.models.StreamTransferRequest`

        :return: 上传成功的结果
        :rtype: :class:``gddriver.models.UploadResult``

        """
        return self._driver.upload_object_via_stream(
            self._connector,
            request=request
        )

    def download_file(self, request):
        """
        下载文件到指定的路径

        :param request: 文件下载请求
        :type request: :class:`gddriver.models.FileTransferRequest`

        """
        return self._driver.download_file(
            self._connector,
            request=request
        )

    def download_object_as_stream(self, request):
        """
        返回一个流的生成器.

        :param request: 流式下载请求
        :type request: :class:`gddriver.models.StreamDownloadRequest`

        :return: A generator of file chunk
        :rtype: :class:``gddriver.models.StreamDownloadResult``

        """
        return self._driver.download_object_as_stream(
            self._connector,
            request=request
        )

    def append_object(self, request):
        """
        将一个流追加上传到指定的object中

        :param request: 追加上传请求
        :type request: :class:`gddriver.models.AppendRequest`

        :rtype: ``gddriver.models.AppendResult``
        """
        return self._driver.append_object(
            self._connector,
            request=request
        )

    def archive_object(self, src_container_name, src_object_name, archive_container_name,
                       archive_object_name, **kwargs):
        """
        从当前容器归档一个文件到目标容器

        :param src_container_name: 原文件所在的容器
        :type  src_container_name: :class:`str`

        :param src_object_name: 原始对象名
        :type src_object_name: `str`

        :param archive_container_name: 目标容器
        :type archive_container_name: `str`

        :param archive_object_name: 目标对象名称
        :type archive_object_name: `str`

        :rtype: `boolean`
        """
        return self._driver.archive_object(
            self._connector,
            src_container_name=src_container_name,
            src_object_name=src_object_name,
            archive_container_name=archive_container_name,
            archive_object_name=archive_object_name,
            **kwargs)

    def delete_object(self, object_name, container_name):
        """
        删除对象

        :param container_name:
        :type container_name: `str`

        :param object_name: Object name.
        :type object_name: `str`
        """
        return self._driver.delete_object(
            self._connector,
            object_name=object_name,
            container_name=container_name
        )

    def restore_object(self, src_container_name, src_object_name):
        """
        :param src_object_name: 原始对象名
        :type src_object_name: `str`

        :param src_container_name: 原文件所在的容器
        :type  src_container_name: :class:`str`

        :rtype: :class:`gddriver.models.RestoreResult`
        """
        return self._driver.restore_object(
            self._connector,
            src_container_name=src_container_name,
            src_object_name=src_object_name
        )

    def move_object(self, src_object_name, dst_object_name, container_name=None, dst_container_name=None):
        """
         移动/重命名
         since gddriver 0.2.2

         :param src_object_name: 原始对象名
         :type src_object_name: `str`

         :param dst_object_name: 目标对象名称
         :type dst_object_name: `str`

         :param container_name: 原文件所在的容器
         :type  container_name: :class:`str`

         :param dst_container_name: 目标容器（默认原始容器）
         :type dst_container_name: `str`

        """
        return self._driver.move_object(
            connection=self._connector,
            src_object_name=src_object_name,
            dst_object_name=dst_object_name,
            container_name=container_name,
            dst_container_name=dst_container_name
        )

    def batch_delete_objects(self, object_name_list, container_name):
        """
        批量删除

        :param object_name_list: 待删除列表
        :type object_name_list: list of str

        :param container_name: 容器名称
        :type  container_name: :class:`str`

        """
        return self._driver.batch_delete_objects(
            self._connector,
            object_name_list=object_name_list,
            container_name=container_name
        )

    def copy_object(self, request):
        """

        从当前Container中拷贝一个Object到当指定的Container

        :param request: 原文件所在的容器(bucket)
        :type  request: :class:``gddriver.models.CopyRequest``

        :return:
        """
        return self._driver.copy_object(
            self._connector,
            request=request
        )

    def get_object_sign_url(self, object_name, container_name=None, **kwargs):
        return self._driver.get_object_sign_url(
            self._connector,
            object_name=object_name,
            container_name=container_name,
            **kwargs
        )

    def init_multipart_upload(self, container_name, object_name, **kwargs):
        """
        初始化分片上传
        since gddriver-0.2.4

        :param object_name:  Object name
        :type object_name: str

        :param container_name: Destination container.
        :type container_name: :class:`str`

        :keyword overwrite: 是否覆盖
        :return: upload_id
        :rtype: str
        """
        return self._driver.init_multipart_upload(
            connection=self._connector,
            container_name=container_name,
            object_name=object_name,
            **kwargs
        )

    def upload_part(self, request):
        """ 上传分片
        since gddriver-0.2.4

        :param request:  request object
        :type request: gddriver.models.PartUploadRequest

        :rtype: gddriver.models.PartUploadResult
        """
        return self._driver.upload_part(
            connection=self._connector,
            request=request
        )

    def complete_multipart_upload(self, request):
        """ 完成分片上传，合并所有分片
        since gddriver-0.2.4

        :param request:  request object
        :type request: gddriver.models.CompleteMultipartUploadRequest

        :rtype: gddriver.models.UploadResult
        """
        self._driver.complete_multipart_upload(
            connection=self._connector,
            request=request
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._connector:
            self._connector.close()

    def __getattribute__(self, item):
        """为operator添加与StorageDriver相同的函数，免去填写connection参数的过程
        当执行自匹配调用时，只接收kwargs参数
        """
        try:
            return super(GenericOperator, self).__getattribute__(item)
        except AttributeError:
            base_attr = self._driver.__getattribute__(item)

            def func(**kwargs):
                return base_attr(connection=self._connector, **kwargs)
            return func
